import re

from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class SesothoGlossMapper:

    gloss_dict = parse_csv(get_full_path(
        'parsers/corpora/main/sesotho/resources/gloss.csv'))

    @classmethod
    def map(cls, gloss):
        gloss = cls.clean_gloss(gloss)
        return cls.gloss_dict.get(gloss, '')

    @classmethod
    def clean_gloss(cls, gloss):
        """Clean a Sesotho gloss."""
        for method in [cls.remove_markers,
                       cls.clean_proper_names_gloss_words,
                       cls.remove_nominal_concord_markers,
                       cls.unify_untranscribed_glosses]:
            gloss = method(gloss)
        return gloss

    @classmethod
    def remove_markers(cls, gloss):
        """Remove noun and verb markers."""
        gloss = cls.remove_noun_markers(gloss)
        gloss = cls.remove_verb_markers(gloss)
        return gloss

    @staticmethod
    def remove_noun_markers(gloss):
        """Remove noun markers."""
        return re.sub(r'[nN]\^(?=\d)', '', gloss)

    @staticmethod
    def remove_verb_markers(gloss):
        """Remove verb markers."""
        return re.sub(r'[vs]\^', '', gloss)

    @staticmethod
    def clean_proper_names_gloss_words(gloss):
        """Clean glosses of proper names.

        In proper names substitute 'n^' marker with 'a_'.
        Lowercase the labels of propernames.
        """
        gloss = re.sub(r'[nN]\^([gG]ame|[nN]ame|[pP]lace|[sS]ong)',
                            r'a_\1', gloss)
        if re.search(r'a_(Game|Name|Place|Song)', gloss):
            gloss = gloss.lower()
        return gloss

    @staticmethod
    def remove_nominal_concord_markers(gloss):
        """Remove markers for nominal concord."""
        match = re.search(r'^(d|lr|obr|or|pn|ps)\d+', gloss)
        if match:
            pos = match.group(1)
            return re.sub(pos, '', gloss)

        return gloss

    @staticmethod
    def unify_untranscribed_glosses(gloss):
        """Unify untranscribed glosses.

        In Sesotho glossing for words which are not understood or
        couldn't be analyzed are marked by 'word' or by 'xxx'. Turn
        both into the standart '???'.
        """
        if gloss == 'word' or gloss == 'xxx':
            return '???'

        return gloss



import re

from acqdiv.parsers.toolbox.cleaners.morpheme_cleaner \
    import ToolboxMorphemeCleaner
from acqdiv.util.csvparser import parse_csv, parse_pos_ud
from acqdiv.util.path import get_full_path


class QaqetPOSMapper:

    pos_dict = parse_csv(get_full_path(
        'parsers/corpora/main/qaqet/resources/pos.csv'))

    pos_ud_dict = parse_pos_ud(get_full_path(
        'parsers/corpora/main/qaqet/resources/pos.csv'))

    @classmethod
    def map(cls, pos, ud=False):

        if pos.startswith('-') or pos.startswith('='):
            return 'sfx'
        elif pos.endswith('-') or pos.endswith('='):
            return 'sfx'
        else:
            pos = cls.clean_pos(pos)

            if ud:
                return cls.pos_ud_dict.get(pos, '')
            else:
                return cls.pos_dict.get(pos, '')

    @classmethod
    def clean_pos(cls, pos):
        pos = cls.unify_unknowns_morpheme(pos)
        pos = ToolboxMorphemeCleaner.remove_morpheme_delimiters(pos)
        return pos

    @classmethod
    def unify_unknowns_morpheme(cls, morpheme):
        unknown_re = re.compile(r'\bx+|\?{2}|\*{3}')
        return unknown_re.sub('???', morpheme)





class Word:
    """Data class representing a word.

    word_language (str): The language of the word.
    word (str): The word itself.
    word_actual (str): The actual form of the word.
    word_target (str): The target form of the word.
    pos (str): The POS tag of the word.
    pos_ud (str): The Universal Dependency POS tag of the word.
    """

    word_language: str
    word: str
    word_actual: str
    word_target: str
    pos: str
    pos_ud: str
    warning: str

    def __init__(self):
        """Initialize the variables representing a word."""
        self.word_language = ''
        self.word = ''
        self.word_actual = ''
        self.word_target = ''
        self.pos = ''
        self.pos_ud = ''
        self.warning = ''



import re

from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.chat.cleaners.utterance_cleaner \
    import CHATUtteranceCleaner
from acqdiv.parsers.corpora.main.english.gloss_mapper \
    import EnglishGlossMapper
from acqdiv.parsers.corpora.main.english.pos_mapper \
    import EnglishPOSMapper


class EnglishManchester1Cleaner(CHATCleaner):

    @classmethod
    def remove_non_words(cls, morph_tier):
        """Remove all non-words from the morphology tier.

        Non-words include:
            end|end („)
            cm|cm (,)
            bq|bq (“)
            eq|eq (”)
        """
        non_words_regex = re.compile(r'end\|end'
                                     r'|cm\|cm'
                                     r'|bq\|bq'
                                     r'|eq\|eq')

        morph_tier = non_words_regex.sub('', morph_tier)
        return CHATUtteranceCleaner.remove_redundant_whitespaces(morph_tier)

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        for cleaning_method in [
                CHATUtteranceCleaner.remove_terminator, cls.remove_non_words,
                CHATUtteranceCleaner.remove_omissions]:
            morph_tier = cleaning_method(morph_tier)

        return morph_tier

    # ---------- morpheme cleaning ----------

    @classmethod
    def clean_gloss(cls, gloss):
        return EnglishGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return EnglishPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return EnglishPOSMapper.map(pos_ud, ud=True)



from acqdiv.parsers.corpora.main.dene.reader import DeneReader
from acqdiv.parsers.metadata.imdi_parser import IMDIParser
from acqdiv.parsers.toolbox.parser import ToolboxParser
from acqdiv.parsers.corpora.main.dene.cleaner import DeneCleaner


class DeneSessionParser(ToolboxParser):

    def get_record_reader(self):
        return DeneReader()

    def get_metadata_reader(self):
        return IMDIParser(self.metadata_path)

    def get_cleaner(self):
        return DeneCleaner()



from acqdiv.parsers.metadata.parser import MetadataParser


class CHATParser(MetadataParser):
    """
    Subclass of metadata.MetadataParser class to deal with CHAT XML metadata extraction.

    Unlike in the IMDI files, a lot of data above session level is not
    specified in the CHATParser/Talkbank XML files, but instead saved in
    per-corpus files (with the extension .cdc). Therefor, the CHATParser class
    contains significantly less functionality than the IMDI class.
    """
    # do we need to pass in the metadata_path?
    # def __init__(self, metadata_path, path):
        # MetadataParser.__init__(self, metadata_path, path)

    def __init__(self, path):
        super().__init__(path)
        self.metadata["participants"] = self.get_participants(self.root)
        self.metadata["comments"] = self.get_comments(self.root)
        # self.metadata["session"] = self.get_session_data()

        # self.metadata["comments"] = self.get_comments(self.root)

    # TODO: where is the get_sessions stuff? in the unifier?


    def get_participants(self, root):
        """
        Get a list of all session participants.

        Iterate through the participants node, building a list of dictionaries.
        In Talkbank XML, data about participants is encoded as attributes of a
        participant node, rather than child nodes as in IMDI. Each participant
        is a dictionary of the form {attr key: attr value, ...}.

        Args:
            root: Root node of the session XML.

        Returns:
            A list of dictionaries containing participant metadata.
        """
        return [self.parse_attrs(p) for p in root.Participants.participant]


    def get_comments(self, root):
        """
        Extract miscellaneous metadata from Talkbank XML.

        In the non-IMDI corpora, all non-participant metadata are stored in
        child nodes of the comment node. This method performs a simple
        dictionary comprehension to extract them.

        Args:
            root: Root node of the session XML.

        Returns:
            A dictionary with metadata. If an error occurs, do nothing.
        """
        #print({c.attrib['type']: str(c) for c in root.comment})
        try:
            return {c.attrib['type']: str(c) for c in root.comment}
        except Exception:
            pass


import re
from acqdiv.parsers.toolbox.cleaners.morpheme_cleaner \
    import ToolboxMorphemeCleaner
from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class KuWaruGlossMapper:

    gloss_dict = parse_csv(get_full_path(
        'parsers/corpora/main/ku_waru/resources/gloss.csv'))

    @classmethod
    def map(cls, gloss):
        return cls.infer_gloss(gloss)

    @classmethod
    def infer_gloss(cls, gloss):
        gloss = ToolboxMorphemeCleaner.clean(gloss)

        gloss = cls.replace_colons(gloss)
        gloss = cls.remove_bt_tp(gloss)
        gloss = cls.replace_many_to_one(gloss)

        number_person_rgx = re.compile(r'[1-3/]+(SG|DU|PL)')

        # case 1: direct mapping
        if gloss in cls.gloss_dict:
            return cls.gloss_dict[gloss]

        # case 2: number-person combinations
        if number_person_rgx.fullmatch(gloss):
            return gloss

        # case 3: lexical gloss
        if gloss.islower():
            return ''

        # case 4: NER
        if gloss in ['PERSON', 'PLACE', 'TRIBE']:
            return ''

        # case 5: multi-category morpheme
        if '.' in gloss:
            categories = gloss.split('.')

            mapped_categories = []

            for category in categories:
                if number_person_rgx.fullmatch(category):
                    mapped_category = category
                elif category.islower():
                    return ''
                else:
                    mapped_category = cls.gloss_dict.get(category, '???')

                mapped_categories.append(mapped_category)

            return '.'.join(mapped_categories)

        # other
        return ''

    @staticmethod
    def replace_colons(gloss):
        """Replace colons by dots.

        Args:
            gloss (str): The gloss.

        Example:
            IMP:2/3DU => IMP.2/3DU
        """
        return gloss.replace(':', '.')

    @staticmethod
    def remove_bt_tp(gloss):
        """Remove TP and BT categories.

        `BT` denotes baby talk
        `TP` denotes Tok Pisin

        Args:
            gloss (str): The gloss.

        Example:
            banana.BT => banana
        """
        return re.sub(r'\.(BT|TP)', '', gloss)

    @staticmethod
    def replace_many_to_one(gloss):
        """Replace multi-word gloss by one-word gloss.

        Args:
            gloss (str): The gloss.

        Current multi-word glosses mapping to one-word glosses are:
            TAG.Q
            that.ABK
            that.ABU
            that.ANA
            that.END
            this.DEF
            this.IP

        Example:
            this.DEF => PROX
        """
        gloss = gloss.replace('TAG.Q', 'Q')
        gloss = re.sub(r'that\.AB[KU]', 'DEM', gloss)
        gloss = re.sub(r'that\.(ANA|END)', 'DIST', gloss)
        gloss = re.sub(r'this\.(DEF|IP)', 'PROX', gloss)

        return gloss



"""Model for the ACQDIV database."""

from sqlalchemy import (Text, Column, Integer, Boolean, ForeignKey)
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class Corpus(Base):
    """Model for the corpus."""
    __tablename__ = 'corpora'

    id = Column(Text, primary_key=True)
    language = Column(Text)
    iso_639_3 = Column(Text)
    glottolog_code = Column(Text)
    owner = Column(Text)
    acronym = Column(Text)
    name = Column(Text)


class Session(Base):
    """Model for the session."""
    __tablename__ = 'sessions'

    id = Column(Integer, primary_key=True)
    corpus = Column(Text, ForeignKey('corpora.id'))
    source_id = Column(Text, nullable=False)
    media_id = Column(Text)
    date = Column(Text)
    duration = Column(Integer)
    # SQLAlchemy relationship definitions:
    speakers = relationship('Speaker', backref='Session')
    utterances = relationship('Utterance', backref='Session')
    words = relationship('Word', backref='Session')
    morphemes = relationship('Morpheme', backref='Session')


class Speaker(Base):
    """Model for the session speaker."""
    __tablename__ = 'speakers'

    id = Column(Integer, primary_key=True)
    session_id_fk = Column(Integer, ForeignKey('sessions.id'))
    uniquespeaker_id_fk = Column(Integer, ForeignKey('uniquespeakers.id'))
    age_raw = Column(Text)
    age = Column(Text)
    age_in_days = Column(Integer)
    role_raw = Column(Text)
    role = Column(Text)
    macrorole = Column(Text)
    languages_spoken = Column(Text)


class UniqueSpeaker(Base):
    """Model for the unique speaker across all corpora."""
    __tablename__ = 'uniquespeakers'

    id = Column(Integer, primary_key=True)
    speaker_label = Column(Text)
    name = Column(Text)
    birthdate = Column(Text)
    gender = Column(Text)
    corpus = Column(Integer, ForeignKey('corpora.id'))


class Utterance(Base):
    """Model for the utterance of a session."""
    __tablename__ = 'utterances'

    id = Column(Integer, primary_key=True)
    session_id_fk = Column(Integer, ForeignKey('sessions.id'))
    source_id = Column(Text)
    speaker_id_fk = Column(Integer, ForeignKey('speakers.id'))
    addressee_id_fk = Column(Integer, ForeignKey('speakers.id'))
    utterance_raw = Column(Text)
    utterance = Column(Text)
    translation = Column(Text)
    morpheme = Column(Text)
    gloss_raw = Column(Text)
    pos_raw = Column(Text)
    sentence_type = Column(Text)
    childdirected = Column(Boolean)
    start = Column(Text)
    end = Column(Text)
    start_raw = Column(Text)
    end_raw = Column(Text)
    comment = Column(Text)
    # SQLAlchemy relationship definitions:
    words = relationship('Word', backref='Utterance')
    morphemes = relationship('Morpheme', backref='Utterance')


class Word(Base):
    """Model for the word of an utterance."""
    __tablename__ = 'words'

    id = Column(Integer, primary_key=True)
    utterance_id_fk = Column(Integer, ForeignKey('utterances.id'))
    language = Column(Text)
    word = Column(Text)
    pos = Column(Text)
    pos_ud = Column(Text)
    word_actual = Column(Text)
    word_target = Column(Text)
    # SQLAlchemy relationship definitions:
    morphemes = relationship('Morpheme', backref='Word')


class Morpheme(Base):
    """Model for the morpheme of an utterance."""
    __tablename__ = 'morphemes'

    id = Column(Integer, primary_key=True)
    utterance_id_fk = Column(Integer, ForeignKey('utterances.id'))
    word_id_fk = Column(Integer, ForeignKey('words.id'))
    language = Column(Text)
    type = Column(Text)
    morpheme = Column(Text)
    gloss_raw = Column(Text)
    gloss = Column(Text)
    pos_raw = Column(Text)
    pos = Column(Text)
    lemma_id = Column(Text)



import re
from acqdiv.util.csvparser import parse_csv, parse_pos_ud
from acqdiv.util.path import get_full_path


class CreePOSMapper:

    pos_dict = parse_csv(get_full_path(
        'parsers/corpora/main/cree/resources/pos.csv'))

    pos_ud_dict = parse_pos_ud(get_full_path(
        'parsers/corpora/main/cree/resources/pos.csv'))

    @classmethod
    def map(cls, pos, ud=False):
        pos = cls.clean_pos(pos)

        if ud:
            return cls.pos_ud_dict.get(pos, '')
        else:
            return cls.pos_dict.get(pos, '')

    @staticmethod
    def uppercase_pos_in_parentheses(pos):
        """Uppercase POS tags in parentheses.

        Parentheses indicate covert grammatical categories.
        """
        pos_in_parentheses_regex = re.compile(r'(\()(\S+)(\))')
        # extract POS in parentheses
        match = pos_in_parentheses_regex.search(pos)
        if not match:
            return pos
        else:
            # replace by uppercased version
            up_pos = match.group(2).upper()
            return pos_in_parentheses_regex.sub(r'\1{}\3'.format(up_pos), pos)

    @classmethod
    def clean_pos(cls, pos):
        return cls.uppercase_pos_in_parentheses(pos)



from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class JapaneseMiyataGloss2SegmentMapper:

    gloss2seg = parse_csv(get_full_path(
        'parsers/corpora/main/japanese_miyata/resources/gloss2segment.csv'))

    @classmethod
    def map(cls, gloss):
        return cls.gloss2seg.get(gloss, '')



import re


class CHATWordCleaner:

    @classmethod
    def clean(cls, word):
        for cleaning_method in [
                cls.remove_form_markers, cls.remove_drawls,
                cls.remove_pauses_within_words, cls.remove_blocking,
                cls.remove_filler]:
            word = cleaning_method(word)

        return word

    @staticmethod
    def remove_form_markers(word):
        """Remove form markers from the word.

        Coding in CHAT: word ending with @.
        The @ and the part after it are removed.
        """
        form_marker_regex = re.compile(r'@.*')
        return form_marker_regex.sub(r'', word)

    @staticmethod
    def remove_drawls(word):
        """Remove drawls in the word.

        Coding in CHAT: : within or after word
        """
        return word.replace(':', '')

    @staticmethod
    def remove_pauses_within_words(word):
        """Remove pauses within the word.

        Coding in CHAT: ^ within word
        """
        pause_regex = re.compile(r'(\S+?)\^')
        return pause_regex.sub(r'\1', word)

    @staticmethod
    def remove_blocking(word):
        """Remove blockings in the word.

        Coding in CHAT: ^ or ≠ at the beginning of the word.
        """
        return word.lstrip('^').lstrip('≠')

    @staticmethod
    def remove_filler(word):
        """Remove filler marker from the word.

        Coding in CHAT: word starts with & or &-
        """
        filler_regex = re.compile(r'&-|&(?!=)(\S+)')
        return filler_regex.sub(r'\1', word)



from acqdiv.parsers.corpora.main.chintang.reader import ChintangReader
from acqdiv.parsers.toolbox.parser import ToolboxParser
from acqdiv.parsers.corpora.main.chintang.imdi_parser \
    import ChintangIMDIParser
from acqdiv.parsers.corpora.main.chintang.cleaner \
    import ChintangCleaner
from acqdiv.parsers.corpora.main.chintang import tc_cleaner

from acqdiv.util.role import RoleMapper
from acqdiv.util.path import get_full_path


class ChintangSessionParser(ToolboxParser):

    role_mapper = RoleMapper(get_full_path(
        'parsers/corpora/main/chintang/resources/speaker_label2macro_role.csv'
    ))

    def get_record_reader(self):
        return ChintangReader()

    def get_metadata_reader(self):
        return ChintangIMDIParser(self.metadata_path)

    def get_cleaner(self):
        return ChintangCleaner()

    def parse(self):
        session = super().parse()
        tc_cleaner.clean(session)
        return session



from acqdiv.parsers.corpus_parser import CorpusParser
from acqdiv.parsers.corpora.main.sesotho.session_parser \
    import SesothoSessionParser


class SesothoCorpusParser(CorpusParser):

    def get_session_parser(self, session_path):
        return SesothoSessionParser(session_path)



"""Abstract class for corpus parsing."""

import glob
from abc import ABC, abstractmethod

from tqdm import tqdm

from acqdiv.model.corpus import Corpus
from acqdiv.util.uniquespeaker import set_unique_speakers
from acqdiv.util.session_duration import extract_duration


class CorpusParser(ABC):
    """Methods for constructing a corpus instance."""

    def __init__(self, cfg, disable_pbar=False):
        """Initialize config.

        Args:
            cfg (dict): Corpus configuration data.
            disable_pbar (bool): Whether the progressbar should be disabled.
        """
        self.cfg = cfg
        self.disable_pbar = disable_pbar
        tqdm.monitor_interval = 0
        self.corpus = Corpus()

    def parse(self):
        """Get a Corpus instance."""
        corpus = self.corpus
        corpus.iso_639_3 = self.cfg['iso639-3']
        corpus.glottolog_code = self.cfg['glottolog_code']
        corpus.corpus = self.cfg['corpus']
        corpus.language = self.cfg['language']
        corpus.owner = self.cfg['owner']
        corpus.acronym = self.cfg['acronym']
        corpus.name = self.cfg['name']
        corpus.sessions = self.iter_sessions()

        return corpus

    @abstractmethod
    def get_session_parser(self, session_path):
        """Get a session parser.

        Returns:
            acqdiv.parsers.session_parser.SessionParser: The session parser.
        """
        pass

    def iter_sessions(self):
        """Iter the sessions of the corpus.

        Yields:
            acqdiv.parsers.SessionParser: The session parser.
        """
        session_paths = sorted(glob.glob(self.cfg['sessions']))

        with tqdm(session_paths, disable=self.disable_pbar) as pbar:

            for session_path in pbar:
                pbar.set_description(session_path)

                session_parser = self.get_session_parser(session_path)

                if session_parser is not None:

                    session = session_parser.parse()

                    # set unique speakers
                    set_unique_speakers(self.corpus.corpus, session.speakers)

                    # add duration
                    session.duration = extract_duration(self.corpus.corpus,
                                                        session.source_id)

                    # ignore sessions with no utterances
                    if len(session.utterances):
                        if self.disable_pbar:
                            print("\t", session_path)

                        yield session



from acqdiv.parsers.metadata.imdi_parser import IMDIParser
import re


class ChintangIMDIParser(IMDIParser):

    def get_participants(self):
        """
        Get a list of session participants.

        Parses the session tree, collecting information about
        all participants in a dictionary.

        Returns:
            A list of dictionaries representing participants. The
            participant's child nodes are keys with their text as values.
            Note: missing values (e.g. birthdate) skipped when not present.
        """
        participants = []
        for actor in self.root.Session.MDGroup.Actors.getchildren():
            participant = {}

            # go through actor nodes, paying special attention to complex nodes
            for actornode in actor.getchildren():
                actortag = actornode.tag.replace("{http://www.mpi.nl/IMDI/Schema/IMDI}", "") # drop the IMDI stuff

                # deal with special nodes first
                # <Languages>: this may contain several languages -> save as list
                if actortag == 'Languages':
                    langlist = []
                    for lang in actornode.xpath("*[local-name() = 'Language']"):
                        if lang.Id.text != None:
                            try:
                                langlist.append(lang.Id.text.split(':')[1])
                            except IndexError:
                                langlist.append(lang.Id.text)
                    if len(langlist) != 0:
                        participant[actortag.lower()] = " ".join(langlist)
                # <Keys>: marks family relations which may be used for role inference when given in relation to the current target child
                elif actortag == 'Keys':
                    # get code for current target child from session name
                    search_tc_session = re.search('^CL(.+?Ch\\d)R\\d+S\\d+$', str(self.root.Session.Name.text))
                    if search_tc_session is not None:
                        current_target_child = search_tc_session.group(1)
                        # find Key where Name = "FSR_" + the code of the current target child
                        for key in actornode.iterfind('Key'):
                            if re.search('FSR_'+current_target_child, key.get('Name')):
                                participant['family_key'] = str(key.text)

                # all other nodes: simply add to dictionary
                else:
                    if not str(actornode.text).strip() == "": # skip empty fields in the IMDI
                        participant[actortag.lower()] = str(actornode.text).strip() # turn booleans into strings

            # check collected participant dictionary, then append to list of all participants
            if not len(participant) == 0:
                # Chintang only: special role inference
                # Default for participant['role'] is the value taken from the field Role; overwrite this by more specific value UNLESS the existing value is "Target Child"

                # (1) Actor/FamilySocialRole is on average more specific than Actor/Role
                if ('familysocialrole' in participant and
                    participant['familysocialrole'] not in ['Unspecified', 'unspecified'] and
                    participant['role'] not in ['target child', 'Target child', 'Target Child']):
                    participant['role'] = participant['familysocialrole']
                # (2) Even better values can sometimes be found in Actor/Keys. This list of nodes has been evaluated above; the relevant value is now in participant['family_key']
                if ('family_key' in participant and
                    participant['family_key'] not in ['None', 'none', 'n/a', '?', '??'] and
                    participant['role'] not in ['target child', 'Target child', 'Target Child']):
                    participant['role'] = participant['family_key']

                participants.append(participant)

        return participants



import os
from shutil import copyfile

os.mkdir('cha')

for folder in os.listdir('.'):
    if os.path.isdir(folder) and folder != 'cha':
        for cha_file in os.listdir(folder):
            cha_path = os.path.join(folder, cha_file)
            copyfile(cha_path, os.path.join('cha', folder + '_' + cha_file))



import re

from acqdiv.parsers.toolbox.readers.reader import ToolboxReader


class RussianReader(ToolboxReader):

    @classmethod
    def get_actual_utterance(cls, rec):
        return rec.get('text', '')

    @classmethod
    def get_speaker_label(cls, rec):
        return rec.get('EUDICOp', '')

    @classmethod
    def get_seg_tier(cls, rec):
        return rec.get('lem', '')

    @classmethod
    def get_gloss_tier(cls, rec):
        return rec.get('mor', '')

    @classmethod
    def get_pos_tier(cls, rec):
        return rec.get('mor', '')

    @staticmethod
    def get_morpheme_type():
        return 'actual'

    # ---------- tier ----------

    @classmethod
    def get_lang_tier(cls, rec):
        return cls.get_pos_tier(rec)

    # ---------- morpheme words ----------

    @classmethod
    def get_morpheme_words(cls, morpheme_tier):
        return morpheme_tier.split()

    @classmethod
    def iter_gloss_pos_words(cls, gloss_pos_tier):
        """Iter gloss and POS tag of a word.

        Tier \mor contains both glosses and POS, # separated by "-" or ":".
        """
        if gloss_pos_tier:
            words = cls.get_morpheme_words(gloss_pos_tier)
            for word in words:
                # 1) If there is no ":" in a word string, gloss and POS are
                # identical (most frequently the case with PCL 'particle').
                if ':' not in word:
                    gloss_word = word
                    pos_word = word
                # 2) Sub-POS are always separated by "-" (e.g. PRO-DEM-NOUN),
                # subglosses are always separated by ":" (e.g. PST:SG:F).
                # What varies, though, is the character that separates POS from
                # glosses in the word: If the POS is V ('verb') or ADJ
                # ('adjective'), the glosses start behind the first "-", e.g.
                # V-PST:SG:F:IRREFL:IPFV -> POS V, gloss PST.SG.F.IRREFL.IPFV
                elif word.startswith('V') or word.startswith('ADJ'):
                    match_verb_adj = re.search('(V|ADJ)-(.*$)', word)
                    if match_verb_adj:
                        gloss_word = match_verb_adj.group(2)
                        pos_word = match_verb_adj.group(1)
                    else:
                        continue
                # 3) For all other POS, the glosses start behind the first ":",
                # e.g. PRO-DEM-NOUN:NOM:SG -> POS PRO.DEM.NOUN, gloss NOM.SG
                else:
                    match_gloss_pos = re.search('(^[^(V|ADJ)].*?):(.*$)', word)
                    if match_gloss_pos:
                        gloss_word = match_gloss_pos.group(2)
                        pos_word = match_gloss_pos.group(1)
                    else:
                        continue

                yield gloss_word, pos_word

    @classmethod
    def get_gloss_words(cls, gloss_tier):
        return [gloss_w for gloss_w, _ in cls.iter_gloss_pos_words(gloss_tier)]

    @classmethod
    def get_pos_words(cls, pos_tier):
        return [pos_w for _, pos_w in cls.iter_gloss_pos_words(pos_tier)]

    # ---------- morpheme ----------

    @classmethod
    def get_morphemes(cls, morpheme_word):
        return [morpheme_word]

    @classmethod
    def get_langs(cls, morpheme_lang_word):
        if 'FOREIGN' in morpheme_lang_word:
            return ['FOREIGN']
        else:
            return ['Russian']



import re

from acqdiv.parsers.chat.readers.reader import CHATReader


class InuktitutReader(CHATReader):
    """Inferences for Inuktitut."""

    def get_start_time(self):
        return self.record.dependent_tiers.get('tim', '')

    def get_end_time(self):
        return ''

    @staticmethod
    def get_actual_alternative(utterance):
        """Get the actual form of alternatives.

        Coding in CHAT: [=? <words>]
        The actual form is the alternative given in brackets.
        """
        replacement_regex = re.compile(r'(?:<.*?>|\S+) \[=\? (.*?)\]')
        return replacement_regex.sub(r'\1', utterance)

    @staticmethod
    def get_target_alternative(utterance):
        """Get the target form of alternatives.

        Coding in CHAT: [=? <words>]
        The target form is the original form.
        """
        # several scoped words
        alternative_regex1 = re.compile(r'<(.*?)> \[=\? .*?\]')
        clean = alternative_regex1.sub(r'\1', utterance)
        # one scoped word
        alternative_regex2 = re.compile(r'(\S+) \[=\? .*?\]')
        return alternative_regex2.sub(r'\1', clean)

    def get_actual_utterance(self):
        """Get the actual form of the utterance.

        Considers alternatives as well.
        """
        utterance = super().get_actual_utterance()
        return self.get_actual_alternative(utterance)

    def get_target_utterance(self):
        """Get the target form of the utterance.

        Considers alternatives as well.
        """
        utterance = super().get_target_utterance()
        return self.get_target_alternative(utterance)

    def get_morph_tier(self):
        return self.record.dependent_tiers.get('xmor', '')

    @staticmethod
    def iter_morphemes(word):
        """Iter POS tags, segments and glosses of a word.

        Morphemes are separated by a '+'. POS tags are on the left
        separated by a '|' from the segment. There may be several POS tags
        all separated by a '|'. The gloss is on the right separated from the
        segment by a '^'.

        Args:
            word (str): A morpheme word.

        Yields:
            tuple: The next POS tag, segment and gloss in the word.
        """
        morpheme_regex = re.compile(r'(.*)\|(.*?)\^(.*)')
        for morpheme in word.split('+'):
            match = morpheme_regex.search(morpheme)
            if match:
                yield match.group(1), match.group(2), match.group(3)
            else:
                yield '', '', ''

    @classmethod
    def get_segments(cls, seg_word):
        return [seg for _, seg, _ in cls.iter_morphemes(seg_word)]

    @classmethod
    def get_glosses(cls, gloss_word):
        return [gloss for _, _, gloss in cls.iter_morphemes(gloss_word)]

    @classmethod
    def get_poses(cls, pos_word):
        return [pos for pos, _, _ in cls.iter_morphemes(pos_word)]

    @staticmethod
    def get_morpheme_language(seg, gloss, pos):
        if seg.endswith('@e'):
            return 'English'
        else:
            return 'Inuktitut'



from acqdiv.parsers.toolbox.readers.reader import ToolboxReader


class DeneReader(ToolboxReader):

    @classmethod
    def get_actual_utterance(cls, rec):
        return rec.get('full', '')

    @classmethod
    def get_seg_tier(cls, rec):
        return rec.get('seg', '')

    @classmethod
    def get_gloss_tier(cls, rec):
        return rec.get('glo', '')

    @classmethod
    def get_sentence_type(cls, rec):
        utt = rec.get('full_orig')
        if utt.endswith('?'):
            return 'question'
        else:
            return 'default'

    @classmethod
    def get_translation(cls, rec):
        return rec.get('eng_u', '')

    @classmethod
    def get_comment(cls, rec):
        return rec.get('com', '')



import re

from acqdiv.parsers.toolbox.cleaners.morpheme_cleaner \
    import ToolboxMorphemeCleaner
from acqdiv.util.timestamp import unify_timestamp


class ToolboxCleaner:

    @staticmethod
    def remove_redundant_whitespaces(string):
        """Remove redundant whitespaces."""
        string = re.sub(r'\s+', ' ', string)
        string = string.strip()
        return string

    @staticmethod
    def cross_clean(rec_dict):
        return rec_dict

    # ---------- utterance ----------

    @staticmethod
    def unify_unknown(utterance):
        return re.sub('xxx?|www|\*{3}', '???', utterance)

    @classmethod
    def clean_utterance(cls, utterance):
        """Clean up corpus-specific utterances.

        Args:
            utterance (str): The raw utterance.

        Returns:
            str: The cleaned utterance.
        """
        return cls.unify_unknown(utterance)

    @classmethod
    def clean_timestamp(cls, timestamp):
        """Clean the time stamp.

        Returns: str
        """
        return unify_timestamp(timestamp)

    # ---------- utterance word ----------

    @classmethod
    def clean_word(cls, word):
        return cls.unify_unknown(word)

    # ---------- morphology tiers ----------

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        """No cleaning per default."""
        return morph_tier

    @classmethod
    def clean_seg_tier(cls, seg_tier):
        """No cleaning per default."""
        return cls.clean_morph_tier(seg_tier)

    @classmethod
    def clean_gloss_tier(cls, gloss_tier):
        """No cleaning per default."""
        return cls.clean_morph_tier(gloss_tier)

    @classmethod
    def clean_pos_tier(cls, pos_tier):
        """No cleaning per default."""
        return cls.clean_morph_tier(pos_tier)

    @classmethod
    def clean_lang_tier(cls, lang_tier):
        """No cleaning per default."""
        return cls.clean_morph_tier(lang_tier)

    # ---------- morpheme words ----------

    @classmethod
    def clean_morpheme_word(cls, morpheme_word):
        """No cleaning per default."""
        return morpheme_word

    @classmethod
    def clean_seg_word(cls, segment_word):
        """No cleaning per default."""
        return cls.clean_morpheme_word(segment_word)

    @classmethod
    def clean_gloss_word(cls, gloss_word):
        """No cleaning per default."""
        return cls.clean_morpheme_word(gloss_word)

    @classmethod
    def clean_pos_word(cls, pos_word):
        """No cleaning per default."""
        return cls.clean_morpheme_word(pos_word)

    @classmethod
    def clean_lang_word(cls, lang_word):
        """No cleaning per default."""
        return cls.clean_morpheme_word(lang_word)

    # ---------- morphemes ----------

    @classmethod
    def clean_morpheme(cls, morpheme):
        """Clean the morpheme."""
        return ToolboxMorphemeCleaner.remove_morpheme_delimiters(morpheme)

    @classmethod
    def clean_seg(cls, segment):
        """Clean the segment."""
        return cls.clean_morpheme(segment)

    @classmethod
    def clean_gloss_raw(cls, gloss):
        """Clean the original gloss."""
        return cls.clean_morpheme(gloss)

    @classmethod
    def clean_gloss(cls, gloss):
        """Map original gloss to ACQDIV gloss."""
        return ''

    @classmethod
    def clean_pos_raw(cls, pos):
        """Clean the original POS tag."""
        return cls.clean_morpheme(pos)

    @classmethod
    def clean_pos(cls, pos):
        """Map original POS tag to ACQDIV POS tag."""
        return ''

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        """Map original POS tag to Universal Dependency POS tag."""
        return ''

    @classmethod
    def clean_lang(cls, lang):
        """Clean the morpheme language."""
        return cls.clean_morpheme(lang)

    @classmethod
    def clean_id(cls, id_):
        """Clean the morpheme ID."""
        return cls.clean_morpheme(id_)



import re
from acqdiv.util.age import get_age_from_birth_date_session_date, get_age_in_days


class ToolboxAgeUpdater:

    age_pattern = re.compile(r".*;.*\..*")
    cleaned_age = re.compile(r'\d{1,2};\d{1,2}\.\d{1,2}')

    @classmethod
    def update(cls, speaker, recording_date):
        """Update age and age_in_days in speaker.

        First attempts to calculate ages from the speaker's birth date and
        the session's recording date. For speakers where this fails, looks for
        speakers that already have a properly formatted age, transfers this age
        from the age_raw column to the age column and calculates
        age_in_days from it.

        Finally, it looks for speakers that only have an age in years and
        does the same.

        Args:
            speaker (acqdiv.model.speaker.Speaker): The speaker.
            recording_date (str): The recording date.
        """

        if speaker.birth_date:
            ages = get_age_from_birth_date_session_date(speaker.birth_date,
                                                        recording_date)
            speaker.age = ages[0]
            speaker.age_in_days = ages[1]

        if not speaker.age and re.fullmatch(cls.age_pattern, speaker.age_raw):
            speaker.age = speaker.age_raw
            speaker.age_in_days = get_age_in_days(speaker.age)

        if ("None" not in speaker.age_raw
                and "Un" not in speaker.age_raw
                and not speaker.age):

            if not cls.cleaned_age.fullmatch(speaker.age_raw):
                try:
                    ages = cls.clean_incomplete_ages(speaker.age_raw)
                    speaker.age = ages[0]
                    speaker.age_in_days = ages[1]
                except ValueError:
                    pass

    @staticmethod
    def clean_incomplete_ages(age):
        """Cleanly formats an age given in terms of only years and months.

        Args:
            age: A string or integer representing the age in years and
                 optionally months of the speaker.

        Returns:
            A string in the form YY;MM.0 where MM is 0 if only years are given.
        """
        if ";" in age:
            parts = age.split(";")
            years = parts[0]
            months = parts[1]
            days = int(years) * 365 + int(months) * 30
            new_age = age + ".0"
            return new_age, days
        else:
            years = age.split('/')[0]
            clean_years = years + ";0.0"
            days = int(years) * 365
            return clean_years, days



from acqdiv.parsers.toolbox.cleaners.imdi_cleaner import IMDICleaner


class RussianIMDICleaner(IMDICleaner):

    @classmethod
    def clean_gender(cls, gender):
        gender = super().clean_gender(gender)
        return gender.title()



from acqdiv.parsers.chat.parser import CHATParser
from acqdiv.parsers.corpora.main.nungon.reader import NungonReader
from acqdiv.parsers.corpora.main.nungon.cleaner import NungonCleaner


class NungonSessionParser(CHATParser):
    @staticmethod
    def get_reader(session_file):
        return NungonReader(session_file)

    @staticmethod
    def get_cleaner():
        return NungonCleaner()



from acqdiv.parsers.toolbox.readers.reader import ToolboxReader
from acqdiv.parsers.metadata.imdi_parser import IMDIParser
from acqdiv.parsers.toolbox.cleaners.cleaner import ToolboxCleaner
from acqdiv.parsers.toolbox.cleaners.imdi_cleaner import IMDICleaner as ICl
from acqdiv.parsers.toolbox.readers.fileparser import ToolboxFileParser
from acqdiv.parsers.session_parser import SessionParser
from acqdiv.parsers.toolbox.readers.age_updater import ToolboxAgeUpdater

from acqdiv.util.role import RoleMapper
from acqdiv.util.alignment import fix_misalignments, align_words_morphemes
from acqdiv.util.childdirectedness import infer_childdirected

from acqdiv.model.session import Session
from acqdiv.model.speaker import Speaker
from acqdiv.model.utterance import Utterance
from acqdiv.model.word import Word
from acqdiv.model.morpheme import Morpheme

import os


class ToolboxParser(SessionParser):
    """Gathers all data for the DB for a given Toolbox session file.

    Uses the ToolboxReader for reading a toolbox file and
    IMDIParser or CHATParser for reading the corresponding metadata file.
    """
    role_mapper = RoleMapper()

    def get_record_reader(self):
        return ToolboxReader()

    def get_metadata_reader(self):
        return IMDIParser(self.metadata_path)

    def get_cleaner(self):
        return ToolboxCleaner()

    def __init__(self, toolbox_path, metadata_path):
        """Get toolbox and metadata readers.

        Args:
            toolbox_path (str): Path to the toolbox file.
            metadata_path (str): Path to the metadata file.
        """
        self.session = Session()

        self.metadata_path = metadata_path
        self.toolbox_path = toolbox_path

        # get record reader
        self.record_reader = self.get_record_reader()
        # get metadata reader
        self.metadata_reader = self.get_metadata_reader()
        # get cleaner
        self.cleaner = self.get_cleaner()

    def parse(self):
        """Get the session instance.

        Returns:
            acqdiv.model.session.Session: The Session instance.
        """
        self.add_session_metadata()
        self.add_speakers()
        self.add_records()

        return self.session

    def get_media_filenames(self):
        md = self.metadata_reader.metadata

        media_filenames = set()

        for mediafile in md['media']['mediafile']:
            link = mediafile['resourcelink']
            filename = os.path.basename(link)
            filename_without_ext = os.path.splitext(filename)[0]
            media_filenames.add(filename_without_ext)

        return ','.join(name for name in media_filenames)

    def add_session_metadata(self):
        """Add the metadata of a session."""
        self.session.source_id = os.path.splitext(
            os.path.basename(self.toolbox_path))[0]
        self.session.date = self.metadata_reader.metadata['session'].get(
            'date', None)
        self.session.media_filename = self.get_media_filenames()

    def add_speakers(self):
        """Add the speakers of a session."""
        for speaker_dict in self.metadata_reader.metadata['participants']:
            speaker = Speaker()
            speaker.birth_date = ICl.clean_date(
                speaker_dict.get('birthdate', ''))
            speaker.code = ICl.clean_label(speaker_dict.get('code', ''))
            speaker.name = ICl.clean_name(speaker_dict.get('name', ''))
            speaker.languages_spoken = speaker_dict.get('languages', '')

            speaker.age_raw = speaker_dict.get('age', '')
            ToolboxAgeUpdater.update(speaker, self.session.date)

            speaker.role_raw = speaker_dict.get('role', '')
            speaker.role = self.role_mapper.role_raw2role(speaker.role_raw)
            speaker.macro_role = self.role_mapper.infer_macro_role(
                speaker.role_raw, speaker.age_in_days, speaker.code)

            speaker.gender_raw = speaker_dict.get('sex', '')
            speaker.gender = ICl.clean_gender(speaker.gender_raw)
            if not speaker.gender:
                speaker.gender = self.role_mapper.role_raw2gender(
                    speaker.role_raw)

            self.session.speakers.append(speaker)

    def add_records(self):
        """Add the utterances to the session."""
        separator = self.record_reader.get_rec_separator()
        toolbox_file = ToolboxFileParser.parse(self.toolbox_path, separator)

        for rec in toolbox_file.records:
            if self.record_reader.is_record(rec):
                self.add_record(rec)

    def add_record(self, rec):
        """Add the utterance to the session."""
        rec = self.cleaner.cross_clean(rec)

        utt = self.add_utterance(rec)

        actual_utterance = self.cleaner.clean_utterance(
            self.record_reader.get_actual_utterance(rec))
        target_utterance = self.cleaner.clean_utterance(
            self.record_reader.get_target_utterance(rec))
        self.add_words(actual_utterance, target_utterance)

        self.add_morphemes(
            self.record_reader.get_seg_tier(rec),
            self.record_reader.get_gloss_tier(rec),
            self.record_reader.get_pos_tier(rec),
            self.record_reader.get_lang_tier(rec),
            self.record_reader.get_id_tier(rec)
        )

        align_words_morphemes(utt)

    @staticmethod
    def _get_speaker(label, speakers):
        for speaker in speakers:
            if speaker.code == label:
                return speaker

        return None

    def add_utterance(self, rec):
        """Add the utterance to the Session instance.

        Args:
            rec (acqdiv.parsers.toolbox.model.record.Record): The record.
        """
        utt = Utterance()
        self.session.utterances.append(utt)

        speaker_label = self.record_reader.get_speaker_label(rec)
        utt.speaker = self._get_speaker(
            speaker_label, self.session.speakers)
        addressee_label = self.record_reader.get_addressee(rec)
        utt.addressee = self._get_speaker(
            addressee_label, self.session.speakers)
        utt.utterance_raw = self.record_reader.get_actual_utterance(rec)
        utt.utterance = self.cleaner.clean_utterance(utt.utterance_raw)
        utt.sentence_type = self.record_reader.get_sentence_type(rec)
        utt.childdirected = self.record_reader.get_childdirected(rec)
        utt.childdirected = infer_childdirected(utt)
        utt.source_id = self.record_reader.get_source_id(rec)
        utt.start_raw = self.record_reader.get_start_raw(rec)
        utt.start = self.cleaner.clean_timestamp(utt.start_raw)
        utt.end_raw = self.record_reader.get_end_raw(rec)
        utt.end = self.cleaner.clean_timestamp(utt.end_raw)
        utt.translation = self.record_reader.get_translation(rec)
        utt.comment = self.record_reader.get_comment(rec)
        utt.morpheme_raw = self.record_reader.get_seg_tier(rec)
        utt.gloss_raw = self.record_reader.get_gloss_tier(rec)
        utt.pos_raw = self.record_reader.get_pos_tier(rec)

        return utt

    def add_words(self, actual_utterance, target_utterance):
        """Get list of words from the utterance.

        Args:
            actual_utterance (str): The clean actual utterance.
            target_utterance (str): The clean target utterance.
        """
        words = self.record_reader.get_words(actual_utterance)
        utterance = self.session.utterances[-1]

        for word in words:
            w = Word()
            utterance.words.append(w)

            word_clean = self.cleaner.clean_word(word)

            w.word = word_clean
            w.word_actual = w.word
            w.word_target = ''
            w.word_language = ''

    def add_morphemes(
            self, seg_tier, gloss_tier, pos_tier, lang_tier, id_tier):
        # clean morphology tiers
        seg_tier = self.cleaner.clean_seg_tier(seg_tier)
        gloss_tier = self.cleaner.clean_gloss_tier(gloss_tier)
        pos_tier = self.cleaner.clean_pos_tier(pos_tier)
        lang_tier = self.cleaner.clean_lang_tier(lang_tier)
        id_tier = self.cleaner.clean_morph_tier(id_tier)

        # get morpheme words from the respective morphology tiers
        wsegs = self.record_reader.get_seg_words(seg_tier)
        wglosses = self.record_reader.get_gloss_words(gloss_tier)
        wposes = self.record_reader.get_pos_words(pos_tier)
        wlangs = self.record_reader.get_lang_words(lang_tier)
        wids = self.record_reader.get_id_words(id_tier)

        # fix misalignments
        if self.record_reader.get_main_morpheme() == 'segment':
            wsegs, wglosses, wposes, wlangs, wids = fix_misalignments(
                [wsegs, wglosses, wposes, wlangs, wids])
        else:
            wglosses, wsegs, wposes, wlangs, wids = fix_misalignments(
                [wglosses, wsegs, wposes, wlangs, wids])

        utt = self.session.utterances[-1]

        # go through all morpheme words
        for wseg, wgloss, wpos, wlang, wid in zip(
                wsegs, wglosses, wposes, wlangs, wids):

            # clean the morpheme words
            cleaned_wseg = self.cleaner.clean_seg_word(wseg)
            cleaned_wgloss = self.cleaner.clean_gloss_word(wgloss)
            cleaned_wpos = self.cleaner.clean_pos_word(wpos)
            cleaned_wlang = self.cleaner.clean_lang_word(wlang)
            cleaned_wid = self.cleaner.clean_morpheme_word(wid)

            # collect morpheme data of a word
            wmorphemes = []

            # get morphemes from the morpheme words
            segments = self.record_reader.get_segs(cleaned_wseg)
            glosses = self.record_reader.get_glosses(cleaned_wgloss)
            poses = self.record_reader.get_poses(cleaned_wpos)
            langs = self.record_reader.get_langs(cleaned_wlang)
            ids = self.record_reader.get_ids(cleaned_wid)

            # fix misalignments
            if self.record_reader.get_main_morpheme() == 'segment':
                segments, glosses, poses, langs, ids = fix_misalignments(
                    [segments, glosses, poses, langs, ids])
            else:
                glosses, segments, poses, langs, ids = fix_misalignments(
                    [glosses, segments, poses, langs, ids])

            # go through morphemes
            for seg, gloss, pos, lang, id_ in zip(
                    segments, glosses, poses, langs, ids):

                m = Morpheme()

                # clean the morphemes
                m.morpheme = self.cleaner.clean_seg(seg)
                m.gloss_raw = self.cleaner.clean_gloss_raw(gloss)
                m.gloss = self.cleaner.clean_gloss(m.gloss_raw)
                m.pos_raw = self.cleaner.clean_pos_raw(pos)
                m.pos = self.cleaner.clean_pos(pos)
                m.pos_ud = self.cleaner.clean_pos_ud(pos)
                m.morpheme_language = self.cleaner.clean_lang(lang)
                m.lemma_id = self.cleaner.clean_id(id_)
                m.warning = ''
                m.type = self.record_reader.get_morpheme_type()

                wmorphemes.append(m)

            utt.morphemes.append(wmorphemes)



import re
from acqdiv.util.age import get_age_from_birth_date_session_date, get_age_in_days


class IndonesianAgeUpdater:

    @classmethod
    def update(cls, speaker, recording_date):
        """Update `age` and `age_in_days` for speaker.

        Args:
            speaker (acqdiv.model.speaker.Speaker): The speaker.
            recording_date (str): The recording date.
        """
        ages = get_age_from_birth_date_session_date(speaker.birth_date,
                                                    recording_date)

        speaker.age = ages[0]
        speaker.age_in_days = ages[1]

        if not speaker.age:
            speaker.age = cls.clean(speaker.age_raw)

        if not speaker.age_in_days:
            speaker.age_in_days = get_age_in_days(speaker.age)

    @staticmethod
    def clean(age_raw):
        """Clean the age in Indonesian.

        Ages are given in the format "P0Y0M0D", where preceding the Y is 
        the number of years, preceding the M is the number of months, 
        and preceding the D is the number of days.

        This function transforms these ages into our target format "YY;MM.DD".
        Sometimes, we only get P0Y0M. In this case, we return "YY;MM.0".

        Args:
            age_raw: The age in the format P0Y0M0D.

        Returns:
            str: The age in the format YY;MM.DD.
        """
        age_raw = age_raw.split('/')[0]
        age = re.match(r'P(\d*)Y(\d*)?M?(\d*)?D?', age_raw)
        if age:
            if age.group(3) != '':
                days = age.group(3)
            else:
                days = "0"
            if age.group(2) != '':
                months = age.group(2)
            else:
                months = "0"
            years = age.group(1)
            return "{0};{1}.{2}".format(years, months, days)
        else:
            return ''



from acqdiv.parsers.toolbox.cleaners.morpheme_cleaner \
    import ToolboxMorphemeCleaner
from acqdiv.util.csvparser import parse_csv, parse_pos_ud
from acqdiv.util.path import get_full_path


class KuWaruPOSMapper:

    pos_dict = parse_csv(get_full_path(
        'parsers/corpora/main/ku_waru/resources/pos.csv'))

    pos_ud_dict = parse_pos_ud(get_full_path(
        'parsers/corpora/main/ku_waru/resources/pos.csv'))

    @classmethod
    def map(cls, pos, ud=False):
        if ud:
            return cls.infer_pos(pos, cls.pos_ud_dict)
        else:
            return cls.infer_pos(pos, cls.pos_dict)

    @classmethod
    def infer_pos(cls, pos, pos_dict):
        if cls.is_suffix(pos):
            return 'sfx'
        elif cls.is_prefix(pos):
            return 'pfx'
        else:
            pos = ToolboxMorphemeCleaner.clean(pos)
            return pos_dict.get(pos, '')

    @staticmethod
    def is_suffix(pos):
        return pos.startswith('-') or pos.startswith('=')

    @staticmethod
    def is_prefix(pos):
        return pos.endswith('-') or pos.endswith('=')



import re

from acqdiv.parsers.toolbox.cleaners.cleaner import ToolboxCleaner
from acqdiv.parsers.corpora.main.qaqet.gloss_mapper import QaqetGlossMapper
from acqdiv.parsers.corpora.main.qaqet.pos_mapper import QaqetPOSMapper


class QaqetCleaner(ToolboxCleaner):

    # ---------- utterance ----------

    @classmethod
    def remove_events_utterance(cls, utterance):
        """Remove events in utterance.

        Events are enclosed in square brackets, e.g.
        [sound], [laugh], [cry], [sings], ...
        """
        # do not match unknowns such as [x]
        event_regex = re.compile(r'\[(?!x+).*\]')
        utterance = event_regex.sub('', utterance)
        return cls.remove_redundant_whitespaces(utterance)

    @classmethod
    def remove_punctuation(cls, utterance):
        punctuation_regex = re.compile(r'[,?.]')
        utterance = punctuation_regex.sub('', utterance)
        return cls.remove_redundant_whitespaces(utterance)

    @classmethod
    def null_untranscribed_utterance(cls, utterance):
        """Null untranscribed utterance.

        Marked as [x+].
        """
        untranscribed_regex = re.compile(r'\[x+\]')
        if untranscribed_regex.fullmatch(utterance):
            return ''

        return utterance

    @classmethod
    def unify_unknowns_utterance(cls, utterance):
        """Unify unknowns.

        Marked as [x+].
        """
        unk_regex = re.compile(r'\[x+\]')
        return unk_regex.sub('???', utterance)

    @classmethod
    def clean_utterance(cls, utterance):
        for cleaning_method in [
            cls.remove_events_utterance,
            cls.remove_punctuation,
            # cls.null_untranscribed_utterance,
            cls.unify_unknowns_utterance
        ]:
            utterance = cleaning_method(utterance)

        return utterance

    # ---------- morphology tier ----------

    @classmethod
    def null_untranscribed_morphology_tier(cls, morph_tier):
        untranscribed_re = re.compile(r'\?\?|\*{3}|x{1,4}')

        if untranscribed_re.fullmatch(morph_tier):
            return ''

        return morph_tier

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        # morph_tier = cls.null_untranscribed_morphology_tier(morph_tier)
        return morph_tier

    @classmethod
    def remove_events_seg_tier(cls, seg_tier):
        events = {
            'action',
            'click',
            'cry',
            'laugh',
            'raises-eyebrows',
            'shakes-head',
            'sings',
            'sneeze',
            'sound',
            'spits'
        }

        for event in events:
            seg_tier = seg_tier.replace(event, '')

        return cls.remove_redundant_whitespaces(seg_tier)

    @classmethod
    def clean_seg_tier(cls, seg_tier):
        return cls.remove_events_seg_tier(seg_tier)

    @classmethod
    def remove_events_gloss_tier(cls, gloss_tier):
        events = {
            'ACTION',
            'CLICK',
            'CRY',
            'LAUGH',
            'SINGS',
            'SNEEZE',
            'SOUND',
            'SPITS'
        }
        for event in events:
            gloss_tier = gloss_tier.replace(event, '')

        return cls.remove_redundant_whitespaces(gloss_tier)

    @classmethod
    def clean_gloss_tier(cls, gloss_tier):
        return cls.remove_events_gloss_tier(gloss_tier)

    @classmethod
    def remove_events_pos_tier(cls, pos_tier):
        events = {
            'ACTION',
            'GESTURE',
            'SOUND'
        }

        for event in events:
            pos_tier = pos_tier.replace(event, '')

        return cls.remove_redundant_whitespaces(pos_tier)

    @classmethod
    def clean_pos_tier(cls, pos_tier):
        return cls.remove_events_pos_tier(pos_tier)

    # ---------- morpheme ----------

    @classmethod
    def clean_lang_word(cls, lang_word):
        if lang_word == '??':
            return ''

        return lang_word

    # ---------- morpheme ----------

    @classmethod
    def clean_gloss(cls, gloss):
        return QaqetGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return QaqetPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return QaqetPOSMapper.map(pos_ud, ud=True)

    @classmethod
    def lang2lang(cls, lang):
        """Map the original language label to ACQDIV label."""
        mapping = {
            'Q': 'Qaqet',
            'TP': 'Tok Pisin',
            'E': 'English',
            'GE': 'German',
            'K': 'Kuanua'
        }

        return mapping.get(lang, 'Qaqet')

    @classmethod
    def clean_lang(cls, lang):
        lang = super().clean_lang(lang)
        return cls.lang2lang(lang)



from typing import Iterable

from acqdiv.model.session import Session


class Corpus:
    """Data class representing a corpus.

    corpus (str): The corpus name.
    language (str): The corpus language.
    iso_639_3 (str): The ISO language code.
    glottolog_code (str): The Glottolog language code.
    owner (str): The name of corpus owner.
    sessions (Iterable[Session]): The sessions of the corpus.
    """

    corpus: str
    language: str
    iso_639_3: str
    glottolog_code: str
    owner: str
    sessions: Iterable[Session]

    def __init__(self):
        """Initialize the variables representing a corpus."""
        self.corpus = ''
        self.language = ''
        self.iso_639_3 = ''
        self.glottolog_code = ''
        self.owner = ''
        self.acronym = ''
        self.name = ''
        self.sessions = []



from acqdiv.util.csvparser import parse_csv, parse_pos_ud
from acqdiv.util.path import get_full_path


class JapaneseMiyataPOSMapper:

    pos_dict = parse_csv(get_full_path(
                    'parsers/corpora/main/japanese_miyata/resources/pos.csv'))

    pos_ud_dict = parse_pos_ud(get_full_path(
                    'parsers/corpora/main/japanese_miyata/resources/pos.csv'))

    @classmethod
    def map(cls, pos, ud=False):
        pos = cls.clean_pos(pos)
        if ud:
            return cls.pos_ud_dict.get(pos, '')
        else:
            return cls.pos_dict.get(pos, '')

    @classmethod
    def replace_colon_by_dot_pos(cls, pos):
        """Replace the colons in the POS tag by a dot."""
        return pos.replace(':', '.')

    @classmethod
    def clean_pos(cls, pos):
        pos = cls.replace_colon_by_dot_pos(pos)
        return pos



import re

from acqdiv.parsers.toolbox.readers.reader import ToolboxReader


class ChintangReader(ToolboxReader):

    @classmethod
    def get_actual_utterance(cls, rec):
        return rec.get('gw', '')

    @classmethod
    def get_seg_tier(cls, rec):
        return rec.get('mph', '')

    @classmethod
    def get_gloss_tier(cls, rec):
        return rec.get('mgl', '')

    @classmethod
    def get_pos_tier(cls, rec):
        return rec.get('ps', '')

    @classmethod
    def get_lang_tier(cls, rec):
        return rec.get('lg', '')

    @classmethod
    def get_id_tier(cls, rec):
        return rec.get('id', '')

    @classmethod
    def get_childdirected(cls, rec):
        for tier in ['TOS', 'tos']:
            if tier in rec:
                tos_raw = rec[tier]
                if 'directed' in tos_raw:
                    if 'child' in tos_raw:
                        return True
                    else:
                        return False

        return None

    @classmethod
    def get_sentence_type(cls, rec):
        # https://github.com/uzling/acqdiv/issues/253
        # \eng: . = default, ? = question, ! = exclamation
        # \nep: । = default, rest identical.
        # Note this is not a "pipe" but the so-called danda at U+0964
        if 'nep' in rec and rec['nep']:
            match_punctuation = re.search('([।?!])$', rec['nep'])
            if match_punctuation is not None:
                sentence_type = None
                if match_punctuation.group(1) == '।':
                    sentence_type = 'default'
                if match_punctuation.group(1) == '?':
                    sentence_type = 'question'
                if match_punctuation.group(1) == '!':
                    sentence_type = 'exclamation'
                return sentence_type
        elif cls.get_translation(rec):
            match_punctuation = re.search('([।?!])$',
                                          cls.get_translation(rec))
            if match_punctuation is not None:
                sentence_type = None
                if match_punctuation.group(1) == '.':
                    sentence_type = 'default'
                if match_punctuation.group(1) == '?':
                    sentence_type = 'question'
                if match_punctuation.group(1) == '!':
                    sentence_type = 'exclamation'
                return sentence_type
        else:
            return ''





class Participant:

    def __init__(self):
        self.code = ''
        self.name = ''
        self.role = ''
        self.language = ''
        self.corpus = ''
        self.age = ''
        self.sex = ''
        self.group = ''
        self.ses = ''
        self.role = ''
        self.education = ''
        self.custom = ''
        self.birth_date = ''



from acqdiv.parsers.corpora.main.chintang.corpus_parser \
    import ChintangCorpusParser
from acqdiv.parsers.corpora.main.cree.corpus_parser \
    import CreeCorpusParser
from acqdiv.parsers.corpora.main.dene.corpus_parser \
    import DeneCorpusParser
from acqdiv.parsers.corpora.main.english.corpus_parser \
    import EnglishCorpusParser
from acqdiv.parsers.corpora.main.indonesian.corpus_parser \
    import IndonesianCorpusParser
from acqdiv.parsers.corpora.main.inuktitut.corpus_parser \
    import InuktitutCorpusParser
from acqdiv.parsers.corpora.main.japanese_miipro.corpus_parser \
    import JapaneseMiiProCorpusParser
from acqdiv.parsers.corpora.main.japanese_miyata.corpus_parser \
    import JapaneseMiyataCorpusParser
from acqdiv.parsers.corpora.main.ku_waru.corpus_parser \
    import KuWaruCorpusParser
from acqdiv.parsers.corpora.main.nungon.corpus_parser \
    import NungonCorpusParser
from acqdiv.parsers.corpora.main.qaqet.corpus_parser \
    import QaqetCorpusParser
from acqdiv.parsers.corpora.main.russian.corpus_parser \
    import RussianCorpusParser
from acqdiv.parsers.corpora.main.sesotho.corpus_parser \
    import SesothoCorpusParser
from acqdiv.parsers.corpora.main.tuatschin.corpus_parser \
    import TuatschinCorpusParser
from acqdiv.parsers.corpora.main.turkish.corpus_parser \
    import TurkishCorpusParser
from acqdiv.parsers.corpora.main.yucatec.corpus_parser \
    import YucatecCorpusParser


class CorpusParserMapper:

    mappings = {
        'Chintang': ChintangCorpusParser,
        'Cree': CreeCorpusParser,
        'Dene': DeneCorpusParser,
        'English_Manchester1': EnglishCorpusParser,
        'Indonesian': IndonesianCorpusParser,
        'Inuktitut': InuktitutCorpusParser,
        'Japanese_MiiPro': JapaneseMiiProCorpusParser,
        'Japanese_Miyata': JapaneseMiyataCorpusParser,
        'Ku_Waru': KuWaruCorpusParser,
        'Nungon': NungonCorpusParser,
        'Qaqet': QaqetCorpusParser,
        'Russian': RussianCorpusParser,
        'Sesotho': SesothoCorpusParser,
        'Tuatschin': TuatschinCorpusParser,
        'Turkish': TurkishCorpusParser,
        'Yucatec': YucatecCorpusParser,
    }

    @staticmethod
    def map(name):
        return CorpusParserMapper.mappings[name]



import re

from acqdiv.parsers.chat.readers.reader import CHATReader


class EnglishManchester1Reader(CHATReader):

    # TODO: move all corrections to cleaner

    @staticmethod
    def correct_speaker_name(name, label, role, target_child_name, pid):
        """Correct speaker name.

        Corrections:
        - if target child = 'Carl' and speaker_label = 'FAT', set name to 'Ian'
        - if label = RAC, set name to Rachel
        - if name = 'Mother', change to [speaker_role of target_child_name]
        - if name = 'Father', change to [speaker_role of target_child_name]
        - if name = 'Grandfather', change to
            [speaker_role of target_child_name]
        - if name missing and label != INV
            set name to [speaker_role of target_child_name]
        - if name == 'CHI' (and PID == '11312/c-00019860-1'):
            set name to 'Liz'
        """
        # order is important!
        if target_child_name == 'Carl' and label == 'FAT':
            return 'Ian'
        elif name in {'Mother', 'Father', 'Grandfather'}:
            return role + ' of ' + target_child_name
        elif name == 'CHI' and pid == '11312/c-00019860-1':
            return 'Liz'
        elif name:
            return name
        else:
            if label == 'RAC':
                return 'Rachel'
            elif label != 'INV':
                return role + ' of ' + target_child_name
            else:
                return name

    def get_speaker_name(self):
        """Get speaker name."""
        name = super().get_speaker_name()
        label = self.get_speaker_label()
        role = self.get_speaker_role()
        target_child_name = self.get_target_child()[1]
        pid = self.chat.pid

        return self.correct_speaker_name(
            name, label, role, target_child_name, pid)

    @staticmethod
    def correct_speaker_label(label):
        """Correct speaker label.

        Corrections:
        - if label = DAD, change to 'FAT'
        """
        if label == 'DAD':
            return 'FAT'
        else:
            return label

    def get_speaker_label(self):
        label = super().get_speaker_label()
        return self.correct_speaker_label(label)

    def get_record_speaker_label(self):
        label = super().get_record_speaker_label()
        return self.correct_speaker_label(label)

    def get_translation(self):
        return self.get_utterance()

    @staticmethod
    def get_word_language(word):
        if word.endswith('@s:fra'):
            return 'French'
        elif word.endswith('@s:ita'):
            return 'Italian'
        else:
            return 'English'

    @staticmethod
    def iter_morphemes(morph_word):
        """Iter morphemes of a word.

        A word consists of word groups in the case of
            - compounds (marker: +)
            - clitics (marker: ~)

        A word group has the following structure:
        prefix#POS|stem&fusionalsuffix-suffix=gloss

        prefix: segment, no gloss (-> assign segment), no POS (-> assign 'pfx')
        stem: segment, gloss (either from '='-part or segment), POS
        suffix: no segment, gloss, no POS (-> assign 'sfx')

        For every component of the compound '=' is prepended to the part (e.g.
        'n|+n|apple+n|tree' -> '=apple', '=tree'). The POS tag of the whole
        compound is removed.

        Returns:
            tuple: (segment, gloss, pos).
        """
        morpheme_regex = re.compile(r'[^#]+#'
                                    r'|[^\-]+'
                                    r'|[\-][^\-]+')

        # split into word groups (in case of compound, clitic) (if applicable)
        word_groups = re.split(r'[+~]', morph_word)

        # check if word is a compound
        if word_groups[0].endswith('|'):
            # remove POS tag of the whole compound
            del word_groups[0]
            is_compound = True
        else:
            is_compound = False

        for word_group in word_groups:

            # get stem gloss and remove it from morpheme word
            match = re.search(r'(.+)=(\S+)$', word_group)
            if match:
                word_group = match.group(1)
                stem_gloss = match.group(2)
            else:
                stem_gloss = ''

            # iter morphemes
            for match in morpheme_regex.finditer(word_group):
                morpheme = match.group()

                # prefix
                if morpheme.endswith('#'):
                    pfx = morpheme.rstrip('#')
                    segment = pfx
                    gloss = segment
                    pos = 'pfx'
                # sfx
                elif morpheme.startswith('-'):
                    sfx = morpheme.lstrip('-')
                    segment = ''
                    gloss = sfx
                    pos = 'sfx'
                # stem
                else:
                    pos, segment = morpheme.split('|')
                    # take gloss from '='-part, otherwise the segment
                    if stem_gloss:
                        gloss = stem_gloss
                    else:
                        gloss = segment

                    # if it is a compound part
                    if is_compound:
                        # prepend '=' to segment
                        segment = '=' + segment

                yield segment, gloss, pos

    @classmethod
    def get_segments(cls, seg_word):
        return [seg for seg, _, _ in cls.iter_morphemes(seg_word)]

    @classmethod
    def get_glosses(cls, gloss_word):
        return [gloss for _, gloss, _ in cls.iter_morphemes(gloss_word)]

    @classmethod
    def get_poses(cls, pos_word):
        return [pos for _, _, pos in cls.iter_morphemes(pos_word)]

    @staticmethod
    def get_morpheme_language(seg, gloss, pos):
        if pos == 'L2':
            return 'FOREIGN'
        else:
            return 'English'



class IndonesianSpeakerLabelCorrector:
    """Class for correcting speaker labels.

    Speaker label `EXP` is replaced based on the speaker's name.
    """
    name2label = {
        'Bety': 'BET',
        'Okki': 'OKK',
        'Liana': 'LIA',
        'Widya': 'WID',
        'Dalan': 'DAL',
        'Yanti': 'YAN',
        'Lanny': 'LAN',
        'Like': 'LIK',
        'Dini': 'DIN',
        'Erni': 'ERN',
        'Uri': 'URI'
    }

    @classmethod
    def correct_rec_label(cls, utt_label):
        """Correct the label in the utterance."""
        if 'EXP' in utt_label:
            return utt_label[3:]
        else:
            return utt_label[0:3]

    @classmethod
    def correct_speaker_label(cls, speaker_label, speaker_name):
        """Correct the label in the speaker data."""
        if speaker_label == 'EXP':
            return cls.name2label[speaker_name]

        return speaker_label



import re


class CHATUtteranceCleaner:
    """Cleaners for CHAT utterances."""

    @classmethod
    def clean(cls, utterance):
        for cleaning_method in [
                cls.remove_terminator,
                cls.unify_untranscribed,
                cls.handle_repetitions,
                cls.remove_events,
                cls.remove_omissions,
                cls.remove_linkers,
                cls.remove_separators,
                cls.remove_ca,
                cls.remove_pauses_between_words,
                cls.remove_scoped_symbols,
                cls.remove_commas,
                # cls.null_untranscribed_utterances,
                cls.null_event_utterances]:
            utterance = cleaning_method(utterance)

        return utterance

    @staticmethod
    def remove_redundant_whitespaces(utterance):
        """Remove redundant whitespaces in utterances.

        Strips multiple whitespaces as well as leading and trailing
        whitespaces. This method is routinely called by various
        cleaning methods.
        """
        whitespace_regex = re.compile(r'\s+')
        return whitespace_regex.sub(' ', utterance).strip(' ')

    @classmethod
    def remove_terminator(cls, utterance):
        """Remove the terminator from the utterance.

        There are 13 different terminators in CHAT. Coding: [+/.!?"]*[!?.]  .
        """
        # postcodes or nothing may follow terminators
        terminator_regex = re.compile(r'[+/.!?"]*[!?.](?=( \[\+|$))')
        clean = terminator_regex.sub('', utterance)
        return cls.remove_redundant_whitespaces(clean)

    # TODO: check for removal

    @staticmethod
    def null_untranscribed_utterances(utterance):
        """Null utterances containing only untranscribed material.

        Note:
            Nulling means here the utterance is returned as an empty string.
        """
        if utterance == '???':
            return ''
        else:
            return utterance

    # TODO: move to word level

    @classmethod
    def null_event_utterances(cls, utterance):
        """Remove zeros coding events.

        CHAT coding: 0
        """
        regex = re.compile(r'\b0\b')
        utterance = regex.sub('', utterance)
        return cls.remove_redundant_whitespaces(utterance)

    @classmethod
    def remove_events(cls, utterance):
        """Remove events from the utterance.

        Coding in CHAT: word starting with &=.
        """
        event_regex = re.compile(r'&=\S+')
        clean = event_regex.sub('', utterance)
        return cls.remove_redundant_whitespaces(clean)

    @staticmethod
    def handle_repetitions(utterance):
        """Write out repeated words in the utterance.

        Words are repeated without modification.

        Coding in CHAT: [x <number>]  .
        """
        repetition_regex = re.compile(
            r'(?:<([^<]*?)>|(\S+))( \[.*?\])? ?\[x (\d+)\]')
        # build cleaned utterance
        clean = ''
        match_end = 0
        for match in repetition_regex.finditer(utterance):
            # add material preceding match
            match_start = match.start()
            clean += utterance[match_end:match_start]

            # check whether it has scope over one or several words
            if match.group(1):
                words = match.group(1)
            else:
                words = match.group(2)

            # append preceding scoped symbol
            if match.group(3):
                words += match.group(3)

            # repeat the word
            repetitions = int(match.group(4))
            clean += ' '.join([words]*repetitions)

            match_end = match.end()

        # add material following last match
        clean += utterance[match_end:]

        if clean:
            return clean
        else:
            return utterance

    @classmethod
    def remove_omissions(cls, utterance):
        """Remove omissions in the utterance.

        Coding in CHAT: word starting with 0.

        Those occurring in square brackets are ignored.
        """
        # if not a null utterance
        if not utterance.startswith('0['):
            omission_regex = re.compile(r'0\S+[^\]](?=\s|$)')
            clean = omission_regex.sub('', utterance)
            return cls.remove_redundant_whitespaces(clean)

        return utterance

    # TODO: move to word level

    @staticmethod
    def unify_untranscribed(utterance):
        """Unify untranscribed material as ???.

        Coding in CHAT: xxx, yyy, www   .

        Note:
            This type of cleaning is done on the utterance level rather than
            on the word level because `null_untranscribed_utterances` depends
            on it.
        """
        untranscribed_regex = re.compile(r'xxx|yyy|www')
        return untranscribed_regex.sub('???', utterance)

    @staticmethod
    def remove_linkers(utterance):
        """Remove linkers from the utterance.

        Coding in CHAT: +["^,+<] (always in the beginning of utterance).
        """
        linker_regex = re.compile(r'^\+["^,+<]')
        return linker_regex.sub('', utterance).lstrip(' ')

    @staticmethod
    def remove_separators(utterance):
        """Remove separators from the utterance.

        Separators are commas, colons or semi-colons which are surrounded
        by whitespaces.
        """
        separator_regex = re.compile(r' [,:;]( )')
        return separator_regex.sub(r'\1', utterance)

    @classmethod
    def remove_ca(cls, utterance):
        """Remove conversation analysis and satellite markers from utterance.

        Note:
            Only four markers (↓↑‡„“”) are attested in the corpora. Only those
            will be checked for removal.
        """
        ca_regex = re.compile(r'[↓↑‡„“”]')
        clean = ca_regex.sub('', utterance)
        return cls.remove_redundant_whitespaces(clean)

    @classmethod
    def remove_pauses_between_words(cls, utterance):
        """Remove pauses between words from the utterance.

        Coding in CHAT: (.), (..), (...)
        """
        pause_regex = re.compile(r'\(\.{1,3}\)')
        clean = pause_regex.sub('', utterance)
        return cls.remove_redundant_whitespaces(clean)

    @classmethod
    def remove_scoped_symbols(cls, utterance):
        """Remove scoped symbols from utterance.

        Coding in CHAT: < > [.*]    .

        Angle and square brackets as well as the content within the square
        brackets is removed.

        Note:
            Scoped symbols can be nested to any depth:
                - word [...][...]
                - <word [...] word> [...]
                - <<word word> [...] word> [...]
        """
        scope_regex = re.compile(r'<|>|\[.*?\]')
        clean = scope_regex.sub('', utterance)
        return cls.remove_redundant_whitespaces(clean)

    @classmethod
    def remove_commas(cls, utterance):
        """Remove commas from utterance."""
        return re.sub(r',', '', utterance)



from acqdiv.parsers.chat.parser import CHATParser
from acqdiv.parsers.corpora.main.english.reader import \
    EnglishManchester1Reader
from acqdiv.parsers.corpora.main.english.cleaner import \
    EnglishManchester1Cleaner


class EnglishManchester1SessionParser(CHATParser):
    @staticmethod
    def get_reader(session_file):
        return EnglishManchester1Reader(session_file)

    @staticmethod
    def get_cleaner():
        return EnglishManchester1Cleaner()



from acqdiv.parsers.corpora.main.russian.reader import RussianReader
from acqdiv.parsers.corpora.main.russian.cleaner import RussianCleaner
from acqdiv.parsers.toolbox.parser import ToolboxParser
from acqdiv.parsers.metadata.imdi_parser import IMDIParser
from acqdiv.parsers.corpora.main.russian.imdi_cleaner \
    import RussianIMDICleaner as ICl
from acqdiv.parsers.toolbox.readers.age_updater import ToolboxAgeUpdater
from acqdiv.parsers.corpora.main.russian import tc_cleaner
from acqdiv.model.speaker import Speaker
from acqdiv.util.role import RoleMapper
from acqdiv.util.path import get_full_path


class RussianSessionParser(ToolboxParser):

    role_mapper = RoleMapper(get_full_path(
        'parsers/corpora/main/russian/resources/speaker_label2macro_role.csv'
    ))

    def get_record_reader(self):
        return RussianReader()

    def get_metadata_reader(self):
        return IMDIParser(self.metadata_path)

    def get_cleaner(self):
        return RussianCleaner()

    def parse(self):
        session = super().parse()
        tc_cleaner.clean(session)
        return session

    def add_speakers(self):
        for speaker_dict in self.metadata_reader.metadata['participants']:
            speaker = Speaker()
            speaker.birth_date = ICl.clean_date(
                speaker_dict.get('birthdate', ''))
            speaker.code = ICl.clean_label(speaker_dict.get('code', ''))
            speaker.role_raw = speaker_dict.get('familysocialrole', '')
            speaker.name = ICl.clean_name(speaker_dict.get('name', ''))
            speaker.languages_spoken = speaker_dict.get('language', '')

            speaker.age_raw = speaker_dict.get('age', '')
            ToolboxAgeUpdater.update(speaker, self.session.date)

            speaker.role_raw = speaker_dict.get('familysocialrole', '')
            speaker.role = self.role_mapper.role_raw2role(speaker.role_raw)
            speaker.macro_role = self.role_mapper.infer_macro_role(
                speaker.role_raw, speaker.age_in_days, speaker.code)

            speaker.gender_raw = speaker_dict.get('sex', '')
            speaker.gender = ICl.clean_gender(speaker.gender_raw)
            if not speaker.gender:
                speaker.gender = self.role_mapper.role_raw2gender(
                    speaker.role_raw)

            self.session.speakers.append(speaker)

    def add_record(self, rec):
        super().add_record(rec)
        self.delete_morphemes()

    def delete_morphemes(self):
        utt = self.session.utterances[-1]
        utt.morpheme_raw = ''
        utt.gloss_raw = ''
        utt.pos_raw = ''
        utt.morphemes = []



import os
import re

from acqdiv.parsers.corpora.main.indonesian.reader import \
    IndonesianReader
from acqdiv.parsers.corpora.main.indonesian.cleaner \
    import IndonesianCleaner
from acqdiv.parsers.corpora.main.indonesian.label_corrector \
    import IndonesianSpeakerLabelCorrector as Lc
from acqdiv.parsers.corpora.main.indonesian.age_updater \
    import IndonesianAgeUpdater
from acqdiv.util.role import RoleMapper
from acqdiv.util.path import get_full_path
from acqdiv.parsers.metadata.chat_parser import CHATParser
from acqdiv.parsers.toolbox.parser import ToolboxParser
from acqdiv.model.speaker import Speaker
from acqdiv.model.word import Word


class IndonesianSessionParser(ToolboxParser):

    role_mapper = RoleMapper(get_full_path(
        'parsers/corpora/main/indonesian/resources/'
        'speaker_label2macro_role.csv'
    ))

    def get_metadata_reader(self):
        return CHATParser(self.metadata_path)

    def add_session_metadata(self):
        self.session.source_id = os.path.splitext(os.path.basename(
            self.toolbox_path))[0]
        metadata = self.metadata_reader.metadata['__attrs__']
        self.session.date = metadata.get('Date', None)

        return self.session

    def add_speakers(self):
        for speaker_dict in self.metadata_reader.metadata['participants']:
            speaker = Speaker()
            speaker.birth_date = speaker_dict.get('birthday', '')
            speaker.code = speaker_dict.get('id', '')
            speaker.name = speaker_dict.get('name', '')
            speaker.code = Lc.correct_speaker_label(speaker.code, speaker.name)
            speaker.languages_spoken = speaker_dict.get('language', '')

            speaker.age_raw = speaker_dict.get('age', '')
            IndonesianAgeUpdater.update(speaker, self.session.date)

            speaker.role_raw = speaker_dict.get('role', '')
            speaker.role = self.role_mapper.role_raw2role(speaker.role_raw)
            speaker.macro_role = self.role_mapper.infer_macro_role(
                speaker.role_raw, speaker.age_in_days, speaker.code)

            speaker.gender_raw = speaker_dict.get('sex', '')
            speaker.gender = speaker.gender_raw.title()
            if not speaker.gender:
                speaker.gender = self.role_mapper.role_raw2gender(
                    speaker.role_raw)

            if self.is_speaker(speaker):
                self.session.speakers.append(speaker)

    def add_utterance(self, rec):
        utt = super().add_utterance(rec)
        speaker_label = self.record_reader.get_speaker_label(rec)
        speaker_label = Lc.correct_rec_label(speaker_label)
        utt.speaker = self._get_speaker(speaker_label, self.session.speakers)

        return utt

    @staticmethod
    def is_speaker(speaker):
        """Check whether the speaker is a real speaker.

        Skip `AUX` participants.

        Args:
            speaker (Speaker): The `Speaker` instance.
        """
        return speaker.code != 'AUX'

    def get_record_reader(self):
        return IndonesianReader()

    def get_cleaner(self):
        return IndonesianCleaner()

    def add_words(self, actual_utterance, target_utterance):
        utt = self.session.utterances[-1]

        for word in self.record_reader.get_words(actual_utterance):
            w = Word()
            utt.words.append(w)

            w.word_language = ''

            # Distinguish between word and word_target;
            # otherwise the target word is identical to the actual word
            if re.search('\(', word):
                w.word_target = re.sub('[()]', '', word)
                w.word = re.sub('\([^)]+\)', '', word)
                w.word_actual = w.word
            else:
                w.word_target = re.sub('xxx?|www', '???', word)
                w.word = re.sub('xxx?', '???', word)
                w.word_actual = w.word



from acqdiv.util.csvparser import parse_csv, parse_pos_ud
from acqdiv.util.path import get_full_path


class InuktitutPOSMapper:

    pos_dict = parse_csv(get_full_path(
        'parsers/corpora/main/inuktitut/resources/pos.csv'))

    pos_ud_dict = parse_pos_ud(get_full_path(
        'parsers/corpora/main/inuktitut/resources/pos.csv'))

    @classmethod
    def map(cls, pos, ud=False):
        pos = cls.clean_pos(pos)

        if ud:
            return cls.pos_ud_dict.get(pos, '')
        else:
            return cls.pos_dict.get(pos, '')

    @classmethod
    def clean_pos(cls, pos):
        """Replace the POS tag separator."""
        return cls.replace_pos_separator(pos)

    @staticmethod
    def replace_pos_separator(pos):
        """Replace the POS tag separator.

        A morpheme may have several POS tags separated by a pipe.
        POS tags to the right are subcategories of the POS tags to the left.
        The separator is replaced by a dot.

        Args:
            pos (str): The POS tag.

        Returns:
            str: POS tag separator replaced by a dot.
        """
        return pos.replace('|', '.')



from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class InuktitutGlossMapper:

    gloss_dict = parse_csv(get_full_path(
        'parsers/corpora/main/inuktitut/resources/gloss.csv'))

    @classmethod
    def map(cls, gloss):
        gloss = cls.clean_gloss(gloss)
        return cls.gloss_dict.get(gloss, '')

    @classmethod
    def clean_gloss(cls, gloss):
        """Replace the stem and grammatical gloss connector."""
        return cls.replace_stem_gram_gloss_connector(gloss)

    @staticmethod
    def replace_stem_gram_gloss_connector(gloss):
        """Replace the stem and grammatical gloss connector.

        A stem gloss is connected with a grammatical gloss by an ampersand.
        The connector is replaced by a dot.

        Args:
            gloss (str): The gloss.

        Returns:
            str: The stem and grammatical connector replaced by a dot.
        """
        return gloss.replace('&', '.')



from configparser import ConfigParser

from acqdiv.util.path import get_full_path
from acqdiv.util.csvparser import parse_csv


def get_roles():
    roles = ConfigParser(delimiters='=')
    roles.optionxform = str
    roles.read(get_full_path('util/resources/role_mapping.ini'))

    return roles


class RoleMapper:
    """Class for mapping various roles to and from other fields."""

    roles = get_roles()

    def __init__(self, path_label2macro_role=None):
        if path_label2macro_role:
            self.label2macro_role = parse_csv(path_label2macro_role)
        else:
            self.label2macro_role = {}

    @classmethod
    def role_raw2role(cls, role_raw):
        """Map the original role to the unified role."""
        if role_raw in cls.roles['role_mapping']:
            role = cls.roles['role_mapping'][role_raw]
            # all unknown's and none's listed in the ini become NULL
            if role != 'Unknown' and role != 'None':
                return role

        return ''

    @classmethod
    def role_raw2gender(cls, role_raw):
        """Map the original role to a gender."""
        return cls.roles['role2gender'].get(role_raw, '')
    
    @classmethod
    def role_raw2macrorole(cls, role_raw):
        """Map the original role to the macro role."""
        return cls.roles['role2macrorole'].get(role_raw, '')

    @classmethod
    def age_in_days2macrorole(cls, age_in_days):
        """Map the age to the macro role."""
        if age_in_days:
            if age_in_days <= 4380:
                return 'Child'
            else:
                return 'Adult'
        
        return ''

    def speaker_label2macrorole(self, speaker_label):
        """Map the speaker label to the macro role."""
        macrorole = self.label2macro_role.get(speaker_label, '')
        # ignore all unknown's in the ini file
        if macrorole != 'Unknown':
            return macrorole
        return ''

    def infer_macro_role(self, role_raw, age_in_days, speaker_label):
        """Infer the macro role.
        
        Inference based on: role, age and speaker label
        """
        macro_role = self.role_raw2macrorole(role_raw)
        
        if macro_role != 'Target_Child':
            macro_role = self.age_in_days2macrorole(age_in_days)

            if not macro_role:
                macro_role = self.role_raw2macrorole(role_raw)
            
            if not macro_role:
                macro_role = self.speaker_label2macrorole(speaker_label)
                
        return macro_role



import re

from acqdiv.parsers.metadata.parser import MetadataParser


class IMDIParser(MetadataParser):
    """ Subclass of metadata.MetadataParser to deal with IMDI metadata (Chintang and Russian via S. Stoll) """

    def __init__(self, path):
        super().__init__(path)
        self.metadata["participants"] = self.get_participants()
        self.metadata["session"] = self.get_session_data()
        self.metadata["project"] = self.get_project_data(self.root)
        self.metadata["media"] = self.get_media_data(self.root)


    def get_participants(self):
        """
        Get a list of session participants.

        Parses the session tree, collecting information about
        all participants in a dictionary.

        Returns:
            A list of dictionaries representing participants. The
            participant's child nodes are keys with their text as values.
            Note: missing values (e.g. birthdate) skipped when not present.
        """
        participants = []
        for actor in self.root.Session.MDGroup.Actors.getchildren():
            participant = {}

            # go through actor nodes, paying special attention to complex nodes
            for actornode in actor.getchildren():
                actortag = actornode.tag.replace("{http://www.mpi.nl/IMDI/Schema/IMDI}", "") # drop the IMDI stuff

                # deal with special nodes first
                # <Languages>: this may contain several languages -> save as list
                if actortag == 'Languages':
                    langlist = []
                    for lang in actornode.xpath("*[local-name() = 'Language']"):
                        if lang.Id.text != None:
                            try:
                                langlist.append(lang.Id.text.split(':')[1])
                            except IndexError:
                                langlist.append(lang.Id.text)
                    if len(langlist) != 0:
                        participant[actortag.lower()] = " ".join(langlist)
                # <Keys>: marks family relations which may be used for role inference when given in relation to the current target child
                elif actortag == 'Keys':
                    # get code for current target child from session name
                    search_tc_session = re.search('^CL(.+?Ch\\d)R\\d+S\\d+$', str(self.root.Session.Name.text))
                    if search_tc_session is not None:
                        current_target_child = search_tc_session.group(1)
                        # find Key where Name = "FSR_" + the code of the current target child
                        for key in actornode.iterfind('Key'):
                            if re.search('FSR_'+current_target_child, key.get('Name')):
                                participant['family_key'] = str(key.text)

                # all other nodes: simply add to dictionary
                else:
                    if not str(actornode.text).strip() == "": # skip empty fields in the IMDI
                        participant[actortag.lower()] = str(actornode.text).strip() # turn booleans into strings

            # check collected participant dictionary, then append to list of all participants
            if not len(participant) == 0:
                participants.append(participant)

        return participants


    def get_session_data(self):
        """
        Get session-level metadata.

        Returns:
            A dictionary of session metadata nodes.
        """
        session = {}
        session['id'] = self.metadata['__attrs__']['Cname']
        session['date'] = self.root.Session.Date.text
        session['genre'] = self.root.Session.MDGroup.Content.Genre.text
        session['location'] = self.get_location(self.root)
        session['situation'] = self.root.Session.Description.text
        # session['duration'] = self.get_session_duration(self.root)
        map(lambda x: x.lower(), session)
        return session


    def get_location(self, root):
        """
        Get metadata about recording location.

        Args:
            root: Root node of the session XML.

        Returns:
            A dictionary containing location metadata.
        """
        location = {}
        for e in root.Session.MDGroup.Location.getchildren():
            t = e.tag.replace("{http://www.mpi.nl/IMDI/Schema/IMDI}", "")
            location[t.lower()] = str(e.text)
        return location


    def get_project_data(self, root):
        """
        Get project-level metadata.

        Args:
            root: Root node of the session XML.

        Returns:
            A dictionary containing location metadata.
        """
        project = {}
        project['name'] = root.Session.MDGroup.Project.Title.text
        project['shortname'] = root.Session.MDGroup.Project.Name.text
        project['id'] = root.Session.MDGroup.Project.Id.text
        project['contact'] = self.get_project_contact(root)
        return project


    def get_project_contact(self, root):
        """
        Get contact data of the project owner.

        Args:
            root: Root node of the session XML.

        Returns:
            A dictionary containing contact data for the project owner.
        """
        contact = {}
        for e in root.Session.MDGroup.Project.Contact.getchildren():
            t = e.tag.replace("{http://www.mpi.nl/IMDI/Schema/IMDI}", "")
            contact[t.lower()] = str(e.text)
        return contact


    def get_media_data(self, root):
        """
        Get data about session multimedia resources.

        If there are audio or video files and positions specified,
        return these data. Otherwise returns an empty dictionary.

        Args:
            root: Root node of the session XML.

        Returns:
            List: The media files and their metadata (like duration etc.).
        """
        media = {
            'mediafile': [],
            'writtenresource': [],
            'anonyms': [],
            'source': []
        }
        for e in root.Session.Resources.getchildren():
            t = e.tag.replace("{http://www.mpi.nl/IMDI/Schema/IMDI}", "")
            media[t.lower()].append(self.get_mediafile_data(e))
        return media


    def get_mediafile_data(self, element):
        """
        Helper method to get metadata for individual multimedia resources.

        Args:
            element: A child node of a media node.

        Returns:
            A dictionary of mediafile data.
        """
        mediafile = {}
        for e in element.getchildren():
            t = e.tag.replace("{http://www.mpi.nl/IMDI/Schema/IMDI}", "")
            mediafile[t.lower()] = str(e.text)
        return mediafile


    # def get_session_duration(self, r):
    #     """
    #     get the duration of a session

    #     Args:
    #         r: root node of imdi file
    #     Returns:
    #         the session duration as an integer
    #     """
    #     try:
    #         duration = r.Session.Resources.Source.Keys.Key.text
    #         return self._format_to_seconds(duration)
    #     except AttributeError:
    #         # print("Error! No Duration extracted.")
    #         return None


    # def _format_to_seconds(self, duration):
    #     """
    #     format a string, representing session duration to seconds
    #     Args:
    #         duration: str, 'hh:mm:ss'
    #     Return:
    #         int
    #     """
    #     dur = duration
    #     h, m, s = dur.split(':')
    #     seconds = 3600*int(h)+60*int(m)+float(s)
    #     return seconds


from pathlib import Path

from acqdiv.parsers.corpus_parser import CorpusParser
from acqdiv.parsers.corpora.main.tuatschin.session_parser \
    import TuatschinSessionParser


class TuatschinCorpusParser(CorpusParser):

    def get_session_parser(self, session_path):
        metadata_filename = Path(session_path).with_suffix('.imdi').name
        metadata_filepath = Path(self.cfg['metadata_dir']) / metadata_filename

        # TODO: remove this check once we have all the metadata
        if metadata_filepath.is_file():
            return TuatschinSessionParser(session_path, str(metadata_filepath))

        return None



import re

from acqdiv.parsers.chat.readers.reader import CHATReader
from acqdiv.parsers.corpora.main.turkish.gloss2segment_mapper \
    import TurkishGloss2SegmentMapper as Mp


class TurkishReader(CHATReader):

    def get_start_time(self):
        """Get the start time.

        It is located on the %tim tier.
        """
        time = self.record.dependent_tiers.get('tim', '')
        if not time:
            return ''
        else:
            time_regex = re.compile(r'([\d:]+)')
            return time_regex.search(time).group()

    def get_end_time(self):
        """Get the end time.

        It is located on the %tim tier and might be missing.
        """
        time = self.record.dependent_tiers.get('tim', '')
        if not time:
            return ''
        else:
            time_regex = re.compile(r'-([\d:]+)')
            match = time_regex.search(time)
            if match:
                return match.group(1)
            else:
                return ''

    def get_morph_tier(self):
        return self.record.dependent_tiers.get('xmor', '')

    @staticmethod
    def get_word_language(word):
        if word.endswith('@s:eng'):
            return 'English'
        elif word.endswith('@s:deu'):
            return 'German'
        elif word.endswith('@s:rus'):
            return 'Russian'
        else:
            return 'Turkish'

    @staticmethod
    def iter_morphemes(word):
        """Iter morphemes of a word.

        Morphemes are separated by dashes.

        Structure: stemPOS:substemPOS|stem-suffixgloss&subsuffixgloss

        Stem: segment, no gloss, POS tag
        Suffix: no segment, gloss, no POS tag (-> assign 'sfx')
        No prefixes, the first morpheme is always the stem.
        """
        morphemes = word.split('-')

        # first morpheme is always stem
        stem = morphemes.pop(0)

        # some morpheme words are malformed, null them
        if '|' not in stem:
            stem_seg = '???'
            stem_pos = '???'
        else:
            stem_pos, stem_seg = stem.split('|')

        yield stem_seg, '', stem_pos

        # iter suffixes
        for suffix in morphemes:
            segment = Mp.map(suffix)
            yield segment, suffix, 'sfx'

    @classmethod
    def get_segments(cls, seg_word):
        return [seg for seg, _, _ in cls.iter_morphemes(seg_word)]

    @classmethod
    def get_glosses(cls, gloss_word):
        return [gloss for _, gloss, _ in cls.iter_morphemes(gloss_word)]

    @classmethod
    def get_poses(cls, pos_word):
        return [pos for _, _, pos in cls.iter_morphemes(pos_word)]


from acqdiv.parsers.toolbox.cleaners.morpheme_cleaner \
    import ToolboxMorphemeCleaner
from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class QaqetGlossMapper:

    gloss_dict = parse_csv(get_full_path(
        'parsers/corpora/main/qaqet/resources/gloss.csv'))

    @classmethod
    def map(cls, gloss):
        gloss = ToolboxMorphemeCleaner.remove_morpheme_delimiters(gloss)
        return cls.infer_gloss(gloss)

    @classmethod
    def infer_gloss(cls, gloss):
        if gloss:
            atms_gloss_raw = gloss.split('.')
            gloss = []
            for atm_gl_raw in atms_gloss_raw:
                if atm_gl_raw not in cls.gloss_dict:
                    atm_gl = '???'
                else:
                    atm_gl = cls.gloss_dict[atm_gl_raw]
                gloss.append(atm_gl)
            # If all atm_poses are '', set to None.
            for atm_gloss in gloss:
                if atm_gloss != '???':
                    gloss = '.'.join(gloss)
                    break
            else:
                gloss = ''
        else:
            gloss = ''

        return gloss



from acqdiv.parsers.chat.parser import CHATParser
from acqdiv.parsers.corpora.main.japanese_miyata.reader import \
    JapaneseMiyataReader
from acqdiv.parsers.corpora.main.japanese_miyata.cleaner import \
    JapaneseMiyataCleaner

from acqdiv.util.role import RoleMapper
from acqdiv.util.path import get_full_path


class JapaneseMiyataSessionParser(CHATParser):

    role_mapper = RoleMapper(get_full_path(
        'parsers/corpora/main/japanese_miyata/resources/'
        'speaker_label2macro_role.csv'
    ))

    @staticmethod
    def get_reader(session_file):
        return JapaneseMiyataReader(session_file)

    @staticmethod
    def get_cleaner():
        return JapaneseMiyataCleaner()



from acqdiv.parsers.chat.parser import CHATParser
from acqdiv.parsers.corpora.main.yucatec.reader import YucatecReader
from acqdiv.parsers.corpora.main.yucatec.cleaner import YucatecCleaner
from acqdiv.parsers.corpora.main.yucatec import tc_cleaner

from acqdiv.util.role import RoleMapper
from acqdiv.util.path import get_full_path


class YucatecSessionParser(CHATParser):

    role_mapper = RoleMapper(get_full_path(
        'parsers/corpora/main/yucatec/resources/'
        'speaker_label2macro_role.csv'
    ))

    @staticmethod
    def get_reader(session_file):
        return YucatecReader(session_file)

    @staticmethod
    def get_cleaner():
        return YucatecCleaner()

    def parse(self):
        session = super().parse()
        tc_cleaner.clean(session)
        return session



from acqdiv.model.uniquespeaker import UniqueSpeaker
from typing import Optional


class Speaker:
    """Data class representing a Speaker in a Session.

    code (str): The label of the speaker.
    name (str): The name of the speaker.
    gender_raw (str): The original gender of the speaker.
    gender (str): The cleaned gender of the speaker.
    birth_date (str): The birth date of the speaker.
    age_raw (str): The original age of the speaker.
    age (str): The age of the speaker in the child language acquisition format.
    age_in_days (str): The age of the speaker in days.
    role_raw (str): The original role of the speaker.
    role (str): The unified role of the speaker.
    macro_role (str): The unified macro role of the speaker.
    languages_spoken: The languages spoken by the speaker.
    """

    uniquespeaker: Optional[UniqueSpeaker]
    code: str
    name: str
    gender_raw: str
    gender: str
    birth_date: str
    age_raw: str
    age: str
    age_in_days: str
    role_raw: str
    role: str
    macro_role: str
    languages_spoken: str

    def __init__(self):
        self.uniquespeaker = None
        self.code = ''
        self.name = ''
        self.gender_raw = ''
        self.gender = ''
        self.birth_date = ''
        self.age_raw = ''
        self.age = ''
        self.age_in_days = ''
        self.role_raw = ''
        self.role = ''
        self.macro_role = ''
        self.languages_spoken = ''



from acqdiv.parsers.corpus_parser import CorpusParser
from acqdiv.parsers.corpora.main.dene.session_parser \
    import DeneSessionParser


class DeneCorpusParser(CorpusParser):

    def get_session_parser(self, session_path):
        temp = session_path.replace(
            self.cfg['paths']['sessions_dir'],
            self.cfg['paths']['metadata_dir'])

        metadata_path = temp.replace('.tbt', '.imdi')

        return DeneSessionParser(session_path, metadata_path)





class Record:

    def __init__(self):
        self.uid = -1
        self.participant_code = ''
        self.utterance = ''
        self.start_time = ''
        self.end_time = ''
        self.dependent_tiers = {}



from acqdiv.util.csvparser import parse_csv
from acqdiv.util.path import get_full_path


class CreeGlossMapper:

    gloss_dict = parse_csv(get_full_path(
        'parsers/corpora/main/cree/resources/gloss.csv'))

    @classmethod
    def map(cls, gloss):
        gloss = cls.clean_gloss(gloss)
        return cls.gloss_dict.get(gloss, '')

    @staticmethod
    def replace_gloss_connector(gloss):
        """Replace the gloss connectors.

        There are three different gloss connectors: '.', '+', ','
        ',' adds an additional specification to a gloss, e.g.
        'p,quest” (question particle)'. '+' and ',' are replaced by a dot.
        """
        return gloss.replace(',', '.').replace('+', '.')

    @classmethod
    def clean_gloss(cls, gloss):
        # gloss = cls.replace_gloss_connector(gloss)
        return gloss



import re

from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.chat.cleaners.utterance_cleaner \
    import CHATUtteranceCleaner
from acqdiv.parsers.corpora.main.inuktitut.gloss_mapper \
    import InuktitutGlossMapper
from acqdiv.parsers.corpora.main.inuktitut.pos_mapper \
    import InuktitutPOSMapper


class InuktitutCleaner(CHATCleaner):

    # ---------- cross cleaning ----------

    @staticmethod
    def add_birth_date(speaker_label, name, birth_date):
        if speaker_label == 'ALI' and name == 'Alec':
            birth_date = '1986-08-25'
        elif speaker_label == 'MAE' and name == 'Mae':
            birth_date = '1986-09-02'
        elif speaker_label == 'SUP' and name == 'Suusi':
            birth_date = '1986-05-17'

        return birth_date

    @classmethod
    def clean_speaker_metadata(
            cls, session_filename, speaker_label, name, role,
            age, gender, language, birth_date, target_child):
        """Add birth dates for ALI, MAE and SUP."""
        birth_date = cls.add_birth_date(speaker_label, name, birth_date)
        return speaker_label, name, role, age, gender, language, birth_date

    # ---------- word cleaning ----------

    @staticmethod
    def remove_dashes(word):
        """Remove dashes before/after xxx."""
        dash_regex = re.compile(r'-?(xxx)-?')
        return dash_regex.sub(r'\1', word)

    @classmethod
    def clean_word(cls, word):
        word = cls.remove_dashes(word)
        return super().clean_word(word)

    # ---------- morphology tier cleaning ----------

    @classmethod
    def clean_morph_tier(cls, xmor):
        """Clean the morphology tier 'xmor'."""
        for cleaning_method in [
                CHATUtteranceCleaner.remove_terminator,
                CHATUtteranceCleaner.null_event_utterances,
                CHATUtteranceCleaner.unify_untranscribed,
                CHATUtteranceCleaner.remove_separators,
                CHATUtteranceCleaner.remove_scoped_symbols,
                # CHATUtteranceCleaner.null_untranscribed_utterances
                ]:
            xmor = cleaning_method(xmor)

        return xmor

    # ---------- morpheme word cleaning ----------

    @classmethod
    def clean_morpheme_word(cls, morpheme_word):
        return CHATUtteranceCleaner.remove_terminator(morpheme_word)

    # ---------- morpheme cleaning ----------

    @staticmethod
    def remove_english_marker(seg):
        """Remove the marker for english words.

        English segments are marked with the form marker '@e'.

        Args:
            seg (str): The segment.

        Returns:
            str: The segment without '@e'.
        """
        english_marker_regex = re.compile(r'(\S+)@e')
        return english_marker_regex.sub(r'\1', seg)

    @classmethod
    def clean_segment(cls, seg):
        """Remove english markers from the segment."""
        return cls.remove_english_marker(seg)

    @classmethod
    def clean_gloss(cls, gloss):
        return InuktitutGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return InuktitutPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return InuktitutPOSMapper.map(pos_ud, ud=True)



"""Abstract class for session parsing."""

from abc import ABC, abstractmethod


class SessionParser(ABC):

    @abstractmethod
    def parse(self):
        """Return an instance of a Session.

        Returns:
            acqdiv.model.session.Session: The Session instance.
        """
        pass



import re
from datetime import datetime
from dateutil.relativedelta import relativedelta


def get_age_from_birth_date_session_date(birthdate, sessiondate):
    """Calculate the age from the birth date and the session date.

    This function attempts to calculate the age of a speaker from their
    birth date and the date the session was recorded. The dates must
    be in the format YYYY(-MM-DD).

    Args:
        birthdate (str): Birth date of the the speaker.
        sessiondate: Date the session was recorded.

    Returns:
        An age in the form YY;MM.DD

    Raises:
        BirthdateError: the speaker's birth date is None or invalid
        SessionDateError: the session recording date is None or invalid
    """
    acc_flag_bd = 0
    acc_flag_sd = 0
    try:
        d1 = datetime.strptime(birthdate, "%Y-%m-%d")
    except:
        try:
            d1 = datetime.strptime(birthdate, "%Y")
            acc_flag_bd = 1
        except Exception as bde:
            return ['', '']

    try:
        d2 = datetime.strptime(sessiondate, "%Y-%m-%d")
    except:
        try:
            d2 = datetime.strptime(sessiondate, "%Y")
            acc_flag_sd = 1
        except:
            try:
                d2 = datetime.strptime(sessiondate, "%Y-%m")
                acc_flag_sd = 2
            except Exception as sde:
                return ['', '']

    diff = relativedelta(d2, d1)
    diff_days = d2 - d1
    if acc_flag_bd != 1 and acc_flag_sd != 1:
        if acc_flag_sd != 2:
            age_cform = "{0};{1}.{2}".format(diff.years, diff.months,
                                             diff.days)
        else:
            age_cform = "{0};{1}.0".format(diff.years, diff.months)
    else:
        age_cform = "{0};0.0".format(diff.years)

    age_days = int(diff_days.days)
    return ([age_cform if age_cform != "0;0.0" else None,
             age_days if age_days != 0 else None])


def get_age_in_days(age):
    """Convert the age to days.

    This function takes a string representing an age in our target format
    of YY;MM.DD and calculates the number of days represented by it.

    Args:
        age (str): The age in the format YY;MM.DD.

    Returns:
        int: The age in days.
    """
    age = re.match(r"(\d*);(\d*).(\d*)", age)
    if age:
        years = int(age.group(1))
        months = int(age.group(2))
        days = int(age.group(3))
        out = years * 365 + months * 30 + days
        return out

    return ''



from acqdiv.parsers.chat.cleaners.utterance_cleaner \
    import CHATUtteranceCleaner

from acqdiv.parsers.chat.cleaners.word_cleaner import CHATWordCleaner
from acqdiv.util.timestamp import unify_timestamp


class CHATCleaner:
    """Default cleaner for CHAT corpora."""

    @staticmethod
    def clean_date(date):
        """Clean the date.

        Prescribed format:
            year-month-day
            YYYY-MM-DD

        Returns: str
        """
        mapping = {'JAN': '01', 'FEB': '02', 'MAR': '03', 'APR': '04',
                   'MAY': '05', 'JUN': '06', 'JUL': '07', 'AUG': '08',
                   'SEP': '09', 'OCT': '10', 'NOV': '11', 'DEC': '12'}
        if not date:
            return ''
        else:
            day, month, year = date.split('-')
            month_clean = mapping[month]
            return '-'.join([year, month_clean, day])

    @staticmethod
    def clean_age(age):
        """Reformat age in CHAT corpora.

        CHAT ages are usually given in the format YY;MM.DD. If month and day is
        missing, it is given in the format 'YY;'. If only the day is missing, it is
        given in the format 'YY;MM.'. Missing values will be converted to 0:
        'YY;0.0' or 'YY:MM.0'.

        Args:
            age (str): The raw age.

        Returns:
            str: The reformatted age.
        """
        if age:
            # month and day is missing
            if age.endswith(';'):
                age += '0.0'
            # day is missing
            elif age.endswith('.'):
                age += '0'

        return age

    @staticmethod
    def clean_gender(gender):
        """Clean the gender.

        Returns: str
        """
        return gender.title()

    @staticmethod
    def clean_name(name):
        """Clean the name.

        Returns: str
        """
        return name

    @staticmethod
    def clean_record_speaker_label(session_filename, speaker_label):
        """Clean the speaker label in a record.

        No cleaning by default.

        Returns:
            str: The cleaned speaker label.
        """
        return speaker_label

    @classmethod
    def clean_utterance(cls, utterance):
        """Clean the utterance.

        Returns: str
        """
        return CHATUtteranceCleaner.clean(utterance)

    @classmethod
    def clean_timestamp(cls, timestamp):
        """Clean the time stamp.

        Returns: str
        """
        return unify_timestamp(timestamp)

    @staticmethod
    def clean_translation(translation):
        """Clean the translation.

        No cleaning by default.

        Returns: str
        """
        return translation

    @classmethod
    def clean_word(cls, word):
        """Clean the word.

        Returns: str
        """
        return CHATWordCleaner.clean(word)

    @staticmethod
    def clean_morph_tier(morph_tier):
        """Clean the morphology tier.

        Default cleaner for the segment, gloss and POS tiers.

        No cleaning by default.

        Returns: str
        """
        return morph_tier

    @classmethod
    def clean_seg_tier(cls, seg_tier):
        """Clean the segment tier.

        Returns: str
        """
        return cls.clean_morph_tier(seg_tier)

    @classmethod
    def clean_gloss_tier(cls, gloss_tier):
        """Clean the gloss tier.

        Returns: str
        """
        return cls.clean_morph_tier(gloss_tier)

    @classmethod
    def clean_pos_tier(cls, pos_tier):
        """Clean the POS tier.

        Returns: str
        """
        return cls.clean_morph_tier(pos_tier)

    @staticmethod
    def clean_session_metadata(session_filename, date, media_filename):
        """Clean across session metadata.

        Mostly used for correcting session dates having the default date
        `1984-01-01`.

        No cleaning by default.

        Args:
            session_filename (str): The name of the session file.
            date (str): The session date.
            media_filename (str): The name of the media file.

        Returns:
            Tuple[str, str]: (date, media filename)
        """
        return date, media_filename

    @staticmethod
    def clean_speaker_metadata(
            session_filename, speaker_label, name, role, age,
            gender, language, birth_date, target_child):
        """Clean across speaker metadata.

        Mostly used for correcting speaker metadata so as to match the speakers
        across different sessions later.

        No cleaning by default.

        Args:
            session_filename (str): The name of the session file.
            speaker_label (str): Code/label of speaker.
            name (str): Full name of speaker.
            role (str): Role of speaker in relation to target child.
            age (str): The age of the speaker at the current session.
            gender (str): The gender of the speaker.
            language (str): The language of the speaker.
            birth_date (str): The birth date of the speaker.
            target_child (Tuple[str, str]): The label and name of the target
                child.

        Returns:
            Tuple[str, str, str, str, str, str, str]:
            (speaker_label, name, role, age, gender, language, birth_date)
        """
        return speaker_label, name, role, age, gender, language, birth_date

    @staticmethod
    def utterance_cross_clean(
            raw_utt, actual_utt, target_utt, seg_tier, gloss_tier, pos_tier):
        """Clean across utterance tiers.

        No cleaning by default.

        Returns:
            Tuple[str, str, str, str, str]:
            (actual_utt, target_utt, seg_tier, gloss_tier, pos_tier)
        """
        return actual_utt, target_utt, seg_tier, gloss_tier, pos_tier

    @staticmethod
    def clean_morpheme_word(morpheme_word):
        """Clean the morpheme word.

        Default cleaner for the segment, gloss and POS words.

        No cleaning by default.

        Returns: str
        """
        return morpheme_word

    @classmethod
    def clean_seg_word(cls, seg_word):
        """Clean the segment word.

        Returns: str
        """
        return cls.clean_morpheme_word(seg_word)

    @classmethod
    def clean_gloss_word(cls, gloss_word):
        """Clean the gloss word.

        Returns: str
        """
        return cls.clean_morpheme_word(gloss_word)

    @classmethod
    def clean_pos_word(cls, pos_word):
        """Clean the POS tag word.

        Returns: str
        """
        return cls.clean_morpheme_word(pos_word)

    @staticmethod
    def clean_morpheme(morpheme):
        """Clean the morpheme.

        Default cleaner for the segment, gloss and POS.

        No cleaning by default.

        Returns: str
        """
        return morpheme

    @classmethod
    def clean_segment(cls, segment):
        """Clean the segment.

        Returns: str
        """
        return cls.clean_morpheme(segment)

    @classmethod
    def clean_gloss_raw(cls, gloss):
        """Clean the gloss.

        Returns: str
        """
        return cls.clean_morpheme(gloss)

    @classmethod
    def clean_gloss(cls, gloss):
        """Map original gloss to ACQDIV gloss."""
        return ''

    @classmethod
    def clean_pos_raw(cls, pos):
        """Clean the POS tag.

        Returns: str
        """
        return cls.clean_morpheme(pos)

    @classmethod
    def clean_pos(cls, pos):
        """Map original POS tag to ACQDIV POS tag."""
        return ''

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        """Map original POS tag to Universal Dependency POS tag."""
        return ''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Main command line interface of the acqdiv package.

Invoking commands:
    acqdiv [OPTIONS] <command> [ARGS]

The following commands are supported:
    - load
"""
import os
import time
import acqdiv
import argparse

from acqdiv.loader import Loader


def load(args):
    """Run the loader."""
    start_time = time.time()
    loader = Loader()

    if args.cfg:
        loader.load(cfg_path=args.cfg)
    else:
        loader.load()

    print("%s seconds --- Finished" % (time.time() - start_time))


def get_cmd_args():
    """Get the command-line arguments."""
    parser = argparse.ArgumentParser(
                description=('Main command line interface of the '
                             'acqdiv package.'))
    subparsers = parser.add_subparsers(
        description=('For more information on the sub-commands, '
                     'run acqdiv <sub-command> -h.'))

    # command 'load'
    parser_load = subparsers.add_parser(
        'load', help='Load the data into the database.',
        description=('The loader parses the corpus files '
                     'and imports the data into a SQLite database. '
                     'By default, only the test files are parsed and '
                     'the data is imported into a test database. '
                     'To run the loader on the full database, '
                     'use the flag -f.'))
    parser_load.add_argument(
        '-c', '--cfg', help="Specify a path to a custom ini file.")

    parser_load.set_defaults(func=load)

    return parser.parse_args()


def main():
    here = os.path.abspath(os.path.dirname(acqdiv.__file__))
    os.chdir(here)
    args = get_cmd_args()
    args.func(args)


if __name__ == '__main__':
    main()



import csv


def parse_csv(path, raw_pos=0, mapped_pos=1):
    morpheme_dict = {}

    with open(path) as csv_file:
        reader = csv.reader(csv_file, delimiter=',', quotechar='"')
        for row in reader:
            raw = row[raw_pos]
            mapped = row[mapped_pos]

            morpheme_dict[raw] = mapped

    return morpheme_dict


def parse_pos_ud(path):
    return parse_csv(path, mapped_pos=2)



import re

from acqdiv.parsers.chat.cleaners.cleaner import CHATCleaner
from acqdiv.parsers.chat.cleaners.utterance_cleaner \
    import CHATUtteranceCleaner
from acqdiv.parsers.corpora.main.turkish.gloss_mapper \
    import TurkishGlossMapper
from acqdiv.parsers.corpora.main.turkish.pos_mapper \
    import TurkishPOSMapper


class TurkishCleaner(CHATCleaner):

    @staticmethod
    def clean_name(name):
        if name == 'Unknown':
            return ''
        else:
            return name

    # ---------- morphology tier cleaning ----------

    @classmethod
    def clean_morph_tier(cls, morph_tier):
        return CHATUtteranceCleaner.remove_terminator(morph_tier)

    # ---------- cross cleaning ----------

    @staticmethod
    def single_morph_word(utterance, morph_tier):
        """Handle complexes consisting of a single morphological word.

        A complex consists of several stems that are either joined by + or _.

        A complex is a single morphological word, if it has a single POS tag.
        The orthographic word will be joined by an underscore. Example:
        POS|stem1_stem2-SFX
        word:
            seg = stem1_stem2   gloss = ??? pos = POS
            seg = ???           gloss = SFX pos = ???
        """
        wwords = utterance.split(' ')
        mwords = morph_tier.split(' ')
        wwords_count = len(wwords)
        mwords_count = len(mwords)

        i = 0
        while i < wwords_count and i < mwords_count:
            mword = mwords[i]
            wword = wwords[i]
            if '_' in mword or '+' in mword:
                if '_' not in wword and '+' not in wword:
                    # check if wword and mword are similar (-> misalignment)
                    if wword[:2] in mword:
                        # check if there is a next word (-> missing join sep)
                        if i + 1 < wwords_count:
                            next_word = wwords[i+1]
                            # check if wword and mword are similar
                            if next_word[:2] in mword:
                                del wwords[i+1]
                                wwords[i] += '_' + next_word
                                wwords_count -= 1
            i += 1

        return ' '.join(wwords), morph_tier

    @staticmethod
    def separate_morph_word(utterance, morph_tier):
        """Handle complexes consisting of separate morphological words.

        A complex consists of several stems that are either joined by + or _.

        A complex consists of two morphological words, if it has separate
        POS tags and suffixes. The orthographic word as well as the
        morphological word is split in this case. The POS tag of the whole
        complex is discarded. Example:
        wholePOS|stem1POS|stem1-STEM1SFX_stem2POS|stem2-STEM2SFX
        word1:
            seg = stem1 gloss = ???         pos = stem1POS
            seg = ???   gloss = STEM1SFX    pos = sfx
        word2:
            seg = stem2 gloss = ???         pos = stem2POS
            seg = ???   gloss = STEM2SFX    pos = sfx
        """
        wwords = utterance.split(' ')
        mwords = morph_tier.split(' ')
        wwords_count = len(wwords)
        mwords_count = len(mwords)

        i = 0
        while i < wwords_count and i < mwords_count:
            # check for double POS tag
            match = re.search(r'\S+?\|(\S+\|.*)', mwords[i])
            if match:
                # discard POS tag of whole complex
                mword = match.group(1)
                # remove old word
                del mwords[i]
                mwords_count -= 1
                # add new words
                for j, w in enumerate(re.split(r'[+_]', mword)):
                    mwords.insert(i+j, w)
                    mwords_count += 1

                # check if utterance word is also joined
                if '_' in wwords[i] or '+' in wwords[i]:
                    # same procedure
                    wword = wwords[i]
                    del wwords[i]
                    wwords_count -= 1
                    for j, w in enumerate(re.split(r'[+_]', wword)):
                        wwords.insert(i + j, w)
                        wwords_count += 1
            i += 1

        return ' '.join(wwords), ' '.join(mwords)

    @classmethod
    def utterance_cross_clean(
            cls, raw_utt, actual_utt, target_utt,
            seg_tier, gloss_tier, pos_tier):
        """Cross clean between word and segment utterances."""
        # which morphology tier does not matter, they are all the same
        mor_tier = seg_tier
        actual_utt, mor_tier = cls.single_morph_word(actual_utt, mor_tier)
        actual_utt, mor_tier = cls.separate_morph_word(actual_utt, mor_tier)
        target_utt, mor_tier = cls.single_morph_word(target_utt, mor_tier)
        target_utt, mor_tier = cls.separate_morph_word(target_utt, mor_tier)

        return actual_utt, target_utt, mor_tier, mor_tier, mor_tier

    # ---------- word cleaning ----------

    @staticmethod
    def replace_plus(unit):
        """Replace plus by an underscore.

        Args:
            unit (str): utterance word or segment
        """
        return unit.replace('+', '_')

    @classmethod
    def clean_word(cls, word):
        word = super().clean_word(word)
        return cls.replace_plus(word)

    # ---------- utterance cleaning ----------

    @staticmethod
    def unify_untranscribed(utterance):
        """Unify untranscribed material as ???.

        Same as super method. Additionally, also unifies more than three `y`s,
        `ww` and `x`.
        """
        untranscribed_regex = re.compile(r'\b((?<!\[)x+|y{3,}|w{2,})\b')
        return untranscribed_regex.sub(r'???', utterance)

    # ---------- morpheme cleaning ----------

    # ---------- segment cleaning ----------

    @classmethod
    def clean_segment(cls, segment):
        return cls.replace_plus(segment)

    # ---------- morpheme cleaning ----------

    @classmethod
    def clean_gloss(cls, gloss):
        return TurkishGlossMapper.map(gloss)

    @classmethod
    def clean_pos(cls, pos):
        return TurkishPOSMapper.map(pos)

    @classmethod
    def clean_pos_ud(cls, pos_ud):
        return TurkishPOSMapper.map(pos_ud, ud=True)



import re

from acqdiv.parsers.chat.model.chat import CHAT
from acqdiv.parsers.chat.model.participant import Participant
from acqdiv.parsers.chat.model.record import Record


class CHATFileParser:
    """Methods for creating a CHAT instance."""

    @classmethod
    def parse(cls, session_file):
        """Get a CHAT instance from a CHAT file.

        Args:
            session_file (file/file-like object): A CHAT file.

        Returns:
            CHAT: The CHAT instance.
        """
        chat = CHAT()
        session = session_file.read()
        cls.add_headers(chat, session)
        cls.add_records(chat, session)

        return chat

    @classmethod
    def add_headers(cls, chat, session):
        """Add all headers to the CHAT instance.

        Args:
            chat (CHAT): The CHAT instance.
            session (str): The CHAT file as a string.
        """
        for metadata_field in cls.iter_metadata_fields(session):
            key, content = cls.get_metadata_field(metadata_field)

            if key == 'ID':
                id_fields = cls.get_id_fields(content)
                code = cls.get_id_code(id_fields)
                participant = chat.participants[code]
                participant.age = cls.get_id_age(id_fields)
                participant.language = cls.get_id_language(id_fields)
                participant.group = cls.get_id_group(id_fields)
                # do not overwrite role from @Participant
                if not participant.role:
                    participant.role = cls.get_id_role(id_fields)
                participant.corpus = cls.get_id_corpus(id_fields)
                participant.custom = cls.get_id_custom(id_fields)
                participant.education = cls.get_id_education(id_fields)
                participant.ses = cls.get_id_ses(id_fields)
                participant.sex = cls.get_id_sex(id_fields)

            elif key == 'Participants':
                for participant in cls.iter_participants(content):
                    p_fields = cls.get_participant_fields(participant)
                    participant = Participant()
                    participant.code = cls.get_participant_id(p_fields)
                    participant.role = cls.get_participant_role(p_fields)
                    participant.name = cls.get_participant_name(p_fields)
                    chat.participants[participant.code] = participant

            elif key == 'Media':
                media_fields = cls.get_media_fields(content)
                chat.media_format = cls.get_media_format(media_fields)
                chat.media_filename = cls.get_media_filename(media_fields)
                chat.media_comment = cls.get_media_comment(media_fields)

            elif key.startswith('Birth of'):
                _, _, participant_id = key.split(' ')
                chat.participants[participant_id].birth_date = content

            else:
                if hasattr(chat, key.lower()):
                    setattr(chat, key.lower(), content)

    @classmethod
    def add_records(cls, chat, session):
        """Add the records to the CHAT instance.

        Args:
            chat (CHAT): The CHAT instance.
            session (str): The CHAT file as a string.
        """
        uid = 0
        for rec_str in cls.iter_records(session):
            rec = Record()
            rec.uid = uid
            main_line = cls.get_mainline(rec_str)
            main_line_fields = cls.get_mainline_fields(main_line)
            rec.participant_code = cls.get_mainline_speaker_id(main_line_fields)
            rec.utterance = cls.get_mainline_utterance(main_line_fields)
            rec.start_time = cls.get_mainline_start_time(main_line_fields)
            rec.end_time = cls.get_mainline_end_time(main_line_fields)

            for dependent_tier in cls.iter_dependent_tiers(rec_str):
                key, content = cls.get_dependent_tier(dependent_tier)
                rec.dependent_tiers[key] = content

            chat.records.append(rec)
            uid += 1

    @staticmethod
    def _replace_line_breaks(session):
        """Remove line breaks within record tiers or metadata fields.

        CHAT inserts a line break and a tab when a tier or field becomes too
        long. The line breaks are replaced by a blank space.

        Args:
            session (str): The session.

        Returns:
            str:  The session without line breaks in tiers and fields.
        """
        return session.replace('\n\t', ' ')

    # ---------- metadata ----------

    @classmethod
    def iter_metadata_fields(cls, session):
        """Iter all metadata fields of a session.

        Metadata fields start with @ followed by the key, colon, tab and its
        content. Line breaks within a field are automatically removed and
        replaced by a blank space.

        Args:
            session (str): The session.

        Yields:
            str: The next metadata field.
        """
        metadata_regex = re.compile(r'@.*?:\t')
        session = cls._replace_line_breaks(session)
        for match in re.finditer(r'[^\n]+', session):
            line = match.group()
            if metadata_regex.search(line):
                yield line

            # metadata section ends
            if line.startswith('*'):
                break

    @staticmethod
    def get_metadata_field(metadata_field):
        """Get the key and content of the metadata field.

        The key is the part within the @ and the colon. The content is the
        part after the tab.

        Args:
            metadata_field (str): The metadata field as returned by
                                  'iter_metadata_fields'.

        Returns:
            tuple: (key, content).
        """
        key, content = metadata_field.split(':\t')
        return key.lstrip('@'), content

    # ---------- @Media ----------

    @staticmethod
    def get_media_fields(media):
        """Get the fields of @Media.

        @Media can consist of three fields:
            - filename (without extension)
            - format (e.g. video)
            - comment (e.g. unlinked)

        The comment is optional and an empty string will be returned
        if missing.

        Args:
            media (str): @Media content.

        Returns:
            tuple: (filename, format, comment).
        """
        fields = media.split(', ')
        # if comment is missing
        if len(fields) == 2:
            return fields[0], fields[1], ''
        else:
            return tuple(fields)

    @staticmethod
    def get_media_filename(media_fields):
        return media_fields[0]

    @staticmethod
    def get_media_format(media_fields):
        return media_fields[1]

    @staticmethod
    def get_media_comment(media_fields):
        return media_fields[2]

    # ---------- @Participants ----------

    @staticmethod
    def iter_participants(participants):
        """Iter participants in @Participants.

        @Participants is a comma-separated list of participants.

        Args:
            participants (str): @Participants content.

        Yields:
            str: The next participant.
        """
        participants_regex = re.compile(r'\s*,\s*')
        for participant in participants_regex.split(participants):
            yield participant

    @staticmethod
    def get_participant_fields(participant):
        """Get the fields of a participant.

        A participant can consist of three fields:
            - ID
            - name
            - role

        Name and role can be missing in which case an empty string is returned.
        If only one field is missing, it is the name.

        Args:
            participant (str): A participant as returned by iter_participants.

        Returns:
            tuple: (label, name, role).
        """
        fields = re.split(r'\s+', participant)
        # name and role is missing
        if len(fields) == 1:
            return fields[0], '', ''
        # name is missing
        if len(fields) == 2:
            return fields[0], '', fields[1]
        else:
            return tuple(fields)

    @staticmethod
    def get_participant_id(participant_fields):
        return participant_fields[0]

    @staticmethod
    def get_participant_name(participant_fields):
        return participant_fields[1]

    @staticmethod
    def get_participant_role(participant_fields):
        return participant_fields[2]

    # ---------- @ID ----------

    @staticmethod
    def get_id_fields(id_field):
        """Get the fields of @ID.

        @ID consists of the following fields: language, corpus, code, age,
        sex, group, SES, role, education, custom.

        Args:
            id_field (str): @ID content.

        Returns:
            tuple: (language, corpus, code, age, sex, group, SES, role,
                    education, custom)
        """
        return tuple(id_field[:-1].split('|'))

    @staticmethod
    def get_id_language(id_fields):
        return id_fields[0]

    @staticmethod
    def get_id_corpus(id_fields):
        return id_fields[1]

    @staticmethod
    def get_id_code(id_fields):
        return id_fields[2]

    @staticmethod
    def get_id_age(id_fields):
        return id_fields[3]

    @staticmethod
    def get_id_sex(id_fields):
        return id_fields[4]

    @staticmethod
    def get_id_group(id_fields):
        return id_fields[5]

    @staticmethod
    def get_id_ses(id_fields):
        return id_fields[6]

    @staticmethod
    def get_id_role(id_fields):
        return id_fields[7]

    @staticmethod
    def get_id_education(id_fields):
        return id_fields[8]

    @staticmethod
    def get_id_custom(id_fields):
        return id_fields[9]

    # ---------- Record ----------

    @classmethod
    def iter_records(cls, session):
        """Yield a record of the session.

        A record starts with '*speaker_label:\t' in CHAT. Line breaks within
        the main line and dependent tiers are automatically removed and
        replaced by a blank space.

        Yields:
            str: The next record.
        """
        session = cls._replace_line_breaks(session)
        rec_regex = re.compile(r'\*[A-Za-z0-9]{2,3}:\t.*?(?=\n\*|\n@End)',
                               flags=re.DOTALL)
        # iter all records
        for match in rec_regex.finditer(session):
            rec = match.group()
            yield rec

    # ---------- Main line ----------

    @classmethod
    def get_mainline(cls, rec):
        """Get the main line of the record.

        Args:
            rec (str): The record.

        Returns:
            str: The main line.
        """
        main_line_regex = re.compile(r'^\*.*')
        return main_line_regex.search(rec).group()

    @staticmethod
    def get_mainline_fields(main_line):
        """Get the fields from the main line.

        The main line consists of the speaker ID, utterance and start and end
        time. The start and end times are optional.

        Args:
            main_line (str): The main line.

        Returns:
            tuple: (speaker ID, utterance, start time, end time).
        """
        main_line_regex = re.compile(
            r'\*([A-Za-z0-9]{2,3}):\t(.*?)(\s*\D?(\d+)(_(\d+))?\D?$|$)')
        match = main_line_regex.search(main_line)
        label = match.group(1)
        utterance = match.group(2)

        if match.group(4):
            start = match.group(4)
        else:
            start = ''

        if match.group(6):
            end = match.group(6)
        else:
            end = ''

        return label, utterance, start, end

    @staticmethod
    def get_mainline_speaker_id(main_line_fields):
        return main_line_fields[0]

    @staticmethod
    def get_mainline_utterance(main_line_fields):
        return main_line_fields[1]

    @staticmethod
    def get_mainline_start_time(main_line_fields):
        return main_line_fields[2]

    @staticmethod
    def get_mainline_end_time(main_line_fields):
        return main_line_fields[3]

    # ---------- dependent tiers ----------

    @staticmethod
    def iter_dependent_tiers(rec):
        """Iter the dependent tiers of a record.

        A dependent tier starts with '%'.

        Args:
            rec (str): The record.

        Yields:
            str: The next dependent tier.
        """
        dependent_tier_regex = re.compile(r'(?<=\n)%.*')
        for dependent_tier in dependent_tier_regex.finditer(rec):
            yield dependent_tier.group()

    @staticmethod
    def get_dependent_tier(dependent_tier):
        """Get the key and content of the dependent tier.

        The key is the part with the % and the colon. The content starts after
        the tab.

        Args:
            dependent_tier (str): The dependent tier as returned by
                                  iter_dependent_tiers.

        Returns:
            tuple: (key, content).
        """
        try:
            key, content = dependent_tier.split(':\t')
        # TODO: delete this once the source data (Inuktitut) is fixed
        except ValueError:
            key, content = dependent_tier.split(': ', maxsplit=1)

        return key.lstrip('%'), content



import re

from acqdiv.parsers.toolbox.readers.reader import ToolboxReader


class KuWaruReader(ToolboxReader):

    @classmethod
    def get_speaker_label(cls, rec):
        return rec.get('ELANParticipant', '')

    @classmethod
    def get_actual_utterance(cls, rec):
        return rec.get('tx', '')

    @classmethod
    def get_sentence_type(cls, rec):
        utterance = cls.get_actual_utterance(rec)

        if utterance.endswith('?'):
            return 'question'
        else:
            return 'default'

    @classmethod
    def get_seg_tier(cls, rec):
        return rec.get('mb', '')

    @classmethod
    def get_gloss_tier(cls, rec):
        return rec.get('ge', '')

    @classmethod
    def get_pos_tier(cls, rec):
        return rec.get('ps', '')

    @classmethod
    def get_translation(cls, rec):
        return rec.get('ft', '')

    @classmethod
    def get_lang_tier(cls, rec):
        return cls.get_pos_tier(rec)

    @classmethod
    def get_langs(cls, morpheme_lang_word):
        langs = []
        for lang in cls.get_poses(morpheme_lang_word):
            if '(TP)' in lang:
                langs.append('Tok Pisin')
            else:
                langs.append('Ku Waru')

        return langs

    @classmethod
    def get_pos_words(cls, pos_tier):
        """Get POS tag words.

        Enhancement of the super parser: Keep word in parentheses together
        with the preceding word. Example: v  (eV) -v:FUT
        """
        word_boundary = re.compile(r'(?<![\-=\s])\s+(?![\-=\s]|\(\S+\))')

        if pos_tier:
            return re.split(word_boundary, pos_tier)
        else:
            return []

    @classmethod
    def get_poses(cls, pos_word):
        """Get POS tags.

        Enhancement of the super parser: Keep morpheme in parentheses together
        with the preceding morpheme. Example: v  (eV) -v:PROG
        """
        morpheme_boundary = re.compile(r'\s+(?!\s|\(\S+\))')

        if pos_word:
            return re.split(morpheme_boundary, pos_word)
        else:
            return []

    @classmethod
    def get_lang_words(cls, morpheme_lang_tier):
        return cls.get_pos_words(morpheme_lang_tier)

    @staticmethod
    def get_morpheme_type():
        return 'actual'



from acqdiv.parsers.corpora.main.ku_waru.reader import KuWaruReader
from acqdiv.parsers.corpora.main.ku_waru.cleaner import KuWaruCleaner
from acqdiv.parsers.metadata.cmdi_parser import CMDIParser
from acqdiv.parsers.toolbox.parser import ToolboxParser


class KuWaruSessionParser(ToolboxParser):

    def get_record_reader(self):
        return KuWaruReader()

    def get_metadata_reader(self):
        return CMDIParser(self.metadata_path)

    def get_cleaner(self):
        return KuWaruCleaner()



import re


def unify_timestamp(timestamp_raw):
    """Unify the time stamp.

    Takes the time position in the session recording in the format
    `HH:MM:SS.mmmm` and returns the equivalent in seconds and
    milliseconds.

    Args:
        timestamp_raw (str): The original timestamp of an utterance.

    Returns:
        str: The timestamp in seconds and milliseconds.
    """
    if not timestamp_raw:
        return ''
    times = re.match(r'(\d+):(\d+):(\d+)\.?(\d+)?', timestamp_raw)
    if times:
        fields = times.lastindex
        if fields == 4:
            seconds = int(
                        times.group(1)) * 3600 \
                      + int(times.group(2)) * 60 \
                      + int(times.group(3))
            msecs = times.group(4)
            return "{0}.{1}".format(seconds, msecs)
        elif fields == 3:
            times = re.match(r'(\d+):(\d+):(\d+)', timestamp_raw)
            if times:
                seconds = int(
                            times.group(1)) * 3600 \
                          + int(times.group(2)) * 60 \
                          + int(times.group(3))
                return "{0}.000".format(seconds)
        else:
            return ''
    else:
        return timestamp_raw



import glob
import os

import acqdiv


def get_path_of_most_recent_database():
    pattern = get_full_path('database/acqdiv_corpus_*.sqlite3')

    for path in sorted(glob.glob(pattern), reverse=True):
        return 'sqlite:///{}'.format(path)

    print('Database not found!')

    return None


def get_acqdiv_path():
    return os.path.abspath(os.path.dirname(acqdiv.__file__))


def get_full_path(path_within_acqdiv_package):
    return os.path.join(get_acqdiv_path(), path_within_acqdiv_package)



from acqdiv.parsers.corpus_parser import CorpusParser
from acqdiv.parsers.corpora.main.yucatec.session_parser \
    import YucatecSessionParser


class YucatecCorpusParser(CorpusParser):

    def get_session_parser(self, session_path):
        return YucatecSessionParser(session_path)



from typing import List

from acqdiv.model.utterance import Utterance
from acqdiv.model.speaker import Speaker


class Session:
    """Data class representing a session.

    source_id (str): The session name.
    date (str): The session date.
    media_filename (str): The media file name of the session.
    duration (str): The duration of the session.
    speakers (List[Speaker]): The session speakers.
    utterances (List[Utterance]): The session utterances.
    """

    source_id: str
    date: str
    media_filename: str
    duration: str
    speakers: List[Speaker]
    utterances: List[Utterance]

    def __init__(self):
        """Initialize variables representing a session."""
        self.source_id = ''
        self.date = ''
        self.media_filename = ''
        self.duration = ''
        self.speakers = []
        self.utterances = []
